###Inkboard

from types import MappingProxyType
from typing import TYPE_CHECKING, Literal, Any, Union
from . import logging as ib_logging

__version__ = "0.2.7"
"inkBoard version"


if TYPE_CHECKING:
    ##These are set in the main() function, so they can actually be imported during runtime too.
    CONFIG_FILE: str
    
    from PythonScreenStackManager.pssm.screen import PSSMScreen as screen
    from PythonScreenStackManager.devices import PSSMdevice as device
    from PythonScreenStackManager.elements import Element
    from inkBoard.configuration.configure import config

    integration_objects: MappingProxyType[Literal["integration_entry"],Any]


def getLogger(name: Union[str,None] = None) -> ib_logging.BaseLogger:
    """Convenience method to get a logger with type hinting for additional levels like verbose.
    
    logging docstr:
    Return a logger with the specified name, creating it if necessary.
    If no name is specified, return the root logger.
    """
    return ib_logging.logging.getLogger(name)


class DomainError(ValueError):
    "The supplied entity is not of a valid domain."
    pass

class Singleton(type):
    """
    Use as metaclass (class Classtype(metaclass=Singleton)).
    Ensures only a single instance of this class can exist, without throwing actual errors. Instead, it simply returns the first define instance.
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]




"Functions to validate a device at runtime"

import logging
from . import basedevice
from .basedevice import Device, FEATURES, InkboardDeviceFeatures

_LOGGER = logging.getLogger(__name__)

##Validate function: use assertions, or at least raise assertion errors when validation fails

def validate_power(device: basedevice.Device):
    assert device.power_off.__module__ == device.reboot.__module__ != __package__, "Device needs to define their own power off and reboot functions for the power feature"

def validate_backlight(device: Device):
    assert isinstance(device.backlight, basedevice.BaseBacklight), "device's backlight should be based on inkBoard's BaseBacklight"

def validate_battery(device: Device):
    assert isinstance(device.battery, basedevice.BaseBattery), "device's backlight should be based on inkBoard's BaseBattery"

def validate_network(device: Device):
    assert isinstance(device.network,basedevice.BaseNetwork), "device's network should be based on inkBoard's BaseNetwork"
    if device.has_feature(FEATURES.FEATURE_CONNECTION):
        assert isinstance(device.network,basedevice.BaseConnectionNetwork), "device's network should be based on inkBoard's BaseConnectionNetwork to use the connection feature"

def validate_rotation(device: Device):
    ##how to?
    return

##More proof feature strings need to be based on constants.
##Just, how to get them from the feature class or vice versa
feature_validators = {
    FEATURES.FEATURE_POWER: validate_power,
    FEATURES.FEATURE_POWER: validate_battery,
    FEATURES.FEATURE_NETWORK: validate_network,
    FEATURES.FEATURE_BACKLIGHT: validate_backlight
}

def validate_device(device: basedevice.Device):
    assert isinstance(device, basedevice.BaseDevice), f"{device} should be based on inkBoard's BaseDevice"

    assert isinstance(device._features, InkboardDeviceFeatures), f"Device features should be an instance of InkBoardDeviceFeatures"

    err = False

    for feature, validator in feature_validators.items():
        if not device.has_feature(feature):
            continue

        try:
            validator(device)
        except AssertionError as exce:
            _LOGGER.error(exce.args[0], exc_info=None)
            err = True

    assert not err, "Device failed validation. See logs for more info"




from typing import TYPE_CHECKING, Union
import tracemalloc
from pathlib import Path
import importlib.util

from . import __version__

if TYPE_CHECKING:
    from PythonScreenStackManager.elements import Element

# ---------------------------------------------------------------------------- #
#                               General constants                              #
# ---------------------------------------------------------------------------- #

DESIGNER_INSTALLED: bool = False
DESIGNER_FOLDER: Union[None,Path] = None

if s := importlib.util.find_spec("inkBoarddesigner"):
    try:
        DESIGNER_FOLDER = Path(s.origin).parent
        DESIGNER_INSTALLED = True
    except:
        pass


FuncExceptions = (TypeError, KeyError, IndexError, OSError, RuntimeError)
"General exceptions to catch when calling functions like update. Usage  in try statements as `except FuncExceptions:`"

RAISE : bool = False
"If true, some errors which are only logged in situations like interaction handling and trigger functions are now raised. Also enables memory allocation tracing."

if RAISE:
    # os.environ["PYTHONTRACEMALLOC"] = "1"
    tracemalloc.start(5)

COMMAND_VERSION = "version"
COMMAND_DESIGNER = "designer"
COMMAND_RUN = "run"
COMMAND_PACK = "pack"
COMMAND_INSTALL = "install"
ARGUMENT_CONFIG = "configuration"
"Argument to use to indicate a config file"

IMPORTER_THREADPOOL = "inkboard-import-threadpool"

INKBOARD_FOLDER = Path(__file__).parent.resolve()
"Absolute path to the folder containing the inkBoard module"

FEATURE_CONNECTION = "FEATURE_CONNECTION"

DEFAULT_CONFIG = "./configuration.yaml"
"The default name to use for the config file"

CONFIG_FILE_TYPES = (
                "yaml",
                "yml"
                    )

INKBOARD_COLORS = {
    "inkboard": (19,54,91), #Prussian Blue
    "inkboard-light": (44,107,176), #Lightened version of Prussian Blue
    "inkboard-dark": (35,31,32), #Dark anthracite color
    "inkboard-gray": (63,59,60), #Dark-ish gray color that just looks nice
    "inkboard-grey": (63,59,60), #Synonym color
    "inkboard-white": (255,255,255) #Simply white but putting it in here for completeness
}

INKBOARD_ICON = INKBOARD_FOLDER / "files/icons/inkboard.ico"

##See https://developers.home-assistant.io/docs/core/entity/weather#forecast-data
##Not included: is_daytime, condition
MDI_FORECAST_ICONS : dict = {
                        "datetime" : None,
                        "cloud_coverage": "mdi:cloud-percent",
                        "humidity": "mdi:water-percent",
                        "apparent_temperature": "mdi:thermometer-lines",
                        "dew_point": "mdi:water-thermometer",
                        "precipitation": "mdi:water",
                        "pressure": "mdi:gauge",
                        "temperature": "mdi:thermometer",
                        "templow": "mdi:thermometer-chevron-down",
                        "wind_gust_speed": "mdi:weather-windy",
                        "wind_speed": "mdi:weather-windy",
                        "precipitation_probability": "mdi:water-percent-alert",
                        "uv_index": "mdi:sun-wireless",
                        "wind_bearing": "mdi:windsock"
                            }
"Dict with default icons to use for forecast data lines"


BASE_RELOAD_MODULES = (
    f"{__package__}.core",
    "custom"
)

FULL_RELOAD_MODULES = [
    "core",
    "configuration",
    "dashboard",
    "platforms",
]

for i, mod in enumerate(FULL_RELOAD_MODULES):
    FULL_RELOAD_MODULES[i] = f"{__package__}.{mod}"
    # NO_RELOAD.append(f"{__base_mod}.{mod}")

##Generally: don't reload pssm, should not change when designing elements or platforms which is what the full reload is mainly meant for.
##Full reload should reload all custom elements, platforms outside basedevice, and reset the screen.
##It's mainly for that, or when making platforms; those may not have a decent ide to work with (like for the kobo)
FULL_RELOAD_MODULES = (*FULL_RELOAD_MODULES, "custom")






import sys
import importlib
import importlib.util
import pkgutil
import inspect
import asyncio
import json

from typing import *
from types import MappingProxyType
from pathlib import Path

import inkBoard
from inkBoard.helpers import classproperty, reload_full_module
import inkBoard.integrations

if TYPE_CHECKING:
    from inkBoard import core as CORE

_LOGGER = inkBoard.getLogger(__name__)

def load_custom_functions(core: "CORE"):
    "Imports the modules from custom functions and constructs the dict holding them."

    ##Move this to loader (where integrationloader will be put too.)
    ##Also check what happens when reloading, without reloading the module
    ##I.e. to test what popping from path_importer_cache does -> nothing I can figure out at least

    try:
        from custom import functions as custom_funcs
    except (ImportError, ModuleNotFoundError):
        _LOGGER.debug("No directory for custom functions found")
        return {}

    _LOGGER.debug("Gathering custom functions")

    funcs = {}

    def log_error(mod_name):
        _LOGGER.error(f"Unable to import custom functions module {mod_name}")

    ##Using walk_packages as it can import any function from a file within a module since those are found too.
    pks_w = pkgutil.walk_packages(custom_funcs.__path__, f"{custom_funcs.__package__}.", onerror=log_error)

    for m in pks_w:
        _LOGGER.verbose(f"Importing custom functions from {m.name}")
        ##HA checks module versions in the manifest with cached values to determine reloads
        if m.name in sys.modules:
            mod = sys.modules[m.name]
            mod = importlib.reload(mod)
        else:
            mod = importlib.import_module(m.name,)

        for name, func in inspect.getmembers(mod, lambda x: 
                                                                    inspect.isfunction(x) 
                                                                    and mod.__name__ in x.__module__
                                                                    and not x.__name__.startswith("_")):
            if name in funcs:
                _LOGGER.warning(f"A custom function with the name {name} is already registered.")
                _LOGGER.debug(f"Duplicate function from {func.__module__}, first function is from {funcs[name].__module__}")
                continue
            funcs[name] = func  
    
    ##When to import: Maybe rewalk config dict after importing this and rebuild?
    ##Maybe nah, just let integrations and stuff handle parsing those themselves.
    ##Seems like this works?

    ##So, for screen: just add every function as shorthand with the custom: prefix
    ##For trigger functions etc., have the integration handle in it

    ##Import after importing modules but before calling setup I think
    ##yeah. Do indeed add imported modules to the core, but directly importing from other things is fine too.

    ##Generally for updating integrations etc: need to stop the problem anyways to install them so eh.
    return funcs

def load_custom_elements(core: "CORE"):
    "Loads elements present in the custom/elements folder, and constructs a dict with them."

    try:
        from custom import elements as custom_elts
    except (ImportError, ModuleNotFoundError):
        _LOGGER.debug("No directory for custom functions found")
        return {}

    _LOGGER.debug("Gathering custom functions")

    elts = {}

    def log_error(mod_name):
        _LOGGER.error(f"Unable to import custom functions module {mod_name}")
    
    ##Using walk_packages as it can import any function from a file within a module since those are found too.
    pks_w = pkgutil.walk_packages(custom_elts.__path__, f"{custom_elts.__package__}.", onerror=log_error)

    for m in pks_w:
        _LOGGER.verbose(f"Importing custom functions from {m.name}")
        ##HA checks module versions in the manifest with cached values to determine reloads
        if m.name in sys.modules:
            mod = sys.modules[m.name]
            mod = importlib.reload(mod)
        else:
            mod = importlib.import_module(m.name,)

        elt_dict = core.util.get_module_elements(mod)
        for name, elt in elt_dict.items():
            if name in elts:
                _LOGGER.warning(f"A custom function with the name {name} is already registered.")
                _LOGGER.debug(f"Duplicate function from {elts.__module__}, first function is from {elts[name].__module__}")
                continue
            elts[name] = elt  

    return elts


class IntegrationLoader:
    "Provides bindings to load inkBoard integrations"

    _integration_keys: dict[str,str] = {}
    _installed_integrations: dict[str,Path] = {}
    _integration_modules: dict[str,str] = {}

    _imported_modules: dict[str,str] = {}
    _loaded_integrations: list[str] = set()

    _reload_imports: bool = False

    @classproperty
    def integration_keys(cls) -> dict[str,str]:
        return cls._integration_keys.copy()

    @classproperty
    def imported_integrations(cls) -> dict[str,Path]:
        "The integrations imported for the loaded config, and the Path pointing to them"
        d = {}
        for integration in cls._imported_modules:
            d[integration] = cls._installed_integrations[integration]
        return d

    @classmethod
    def get_integrations(cls, folders: dict[str,Path]):
        "Retrieves all the available integrations in the config's custom folder and inkBoard itself."
        #folders: dict with base module name mapped to the folder path
        cls._reset()


        for base_module, folder in folders.items():
            if folder.exists():
                cls._read_out_folder(base_module, folder)

        return MappingProxyType(cls._installed_integrations)
    
    @classmethod
    def add_integration_config_key(cls, key: str, module_name: str):
        if key in cls._integration_keys:
            int_mod = cls._integration_keys[key]
            _LOGGER.info(f"{key} is already used for a the config of a different integration: {int_mod}")
        else:
            cls._integration_keys[key] = module_name
        return

    @classmethod
    def _read_out_folder(cls, base_module: str, folder: Path):
        folders = folder.iterdir()
        for int_dir in filter(lambda p: (p.is_dir() and not p.name.startswith("_")), folders):
            if int_dir.name in cls._installed_integrations:
                _LOGGER.info(f"Integration {int_dir.name} has already been found in module {cls._integration_modules[int_dir.name]}. Will not import from {base_module}")
                continue

            manifest = int_dir / "manifest.json"
            if not manifest.exists():
                _LOGGER.error(f"Integration folder {int_dir} is missing the manifest.json file.")
                continue

            with open(manifest) as f:
                manifest = json.load(f)
                ##Support for requirements has not yet been implemented
            
            if c := manifest.get("config_entry",False):
                #Will require config_keys, similar to esphome, which can be left empty if needed.
                name = f"{base_module}.{int_dir.name}"
                ##These should not be checked by config key (i.e., save the key in the dict entry); key should be the folder name.
                cls.add_integration_config_key(c, name)
                cls._installed_integrations[int_dir.name] = int_dir
                cls._integration_modules[int_dir.name] = name
            else:
                _LOGGER.error(f"Integrations are required to have a config_entry key {int_dir.name} does not")
                continue

    @classmethod
    def import_integrations(cls, core: "CORE", progress_func=None, value_range=()):
        config = core.config
        import_set: set[tuple[str, Path]] = set()

        ##Import custom via appending to path.
        ##And I think just import via here, but make the set or something recognise the base package too.
        ##Or save strings to the modules instead of their folders.

        for config_entry in cls._integration_keys:
            if config_entry in config.configuration:
                import_set.add((config_entry, cls._integration_keys[config_entry]))
        
        if not import_set: 
            return

        if progress_func:
            progress_func(value_range[0] + 1, f"Importing {len(import_set)} integrations")
            step = int((value_range[1] - value_range[0] - 1)/len(import_set))
            progress = value_range[0] + 1

        for (config_key, name) in import_set:
            integration = name.split(".")[-1]
            if progress_func:
                progress = progress + step
                progress_func(progress, f"Importing integration {integration}")
            module = cls._import_integration(name)
            if not module:
                _LOGGER.error(f"Unable to successfully import integration {integration} from {name}")
            else:
                cls._imported_modules[integration] = module
        
        cls._reload_imports = False
        return
        ##See code in hass core: https://github.com/home-assistant/core/blob/ab5ddb8edfb72d5f5915574f642eba93afc5abdc/homeassistant/loader.py#L1669

    @classmethod
    def _import_integration(cls, name):
            
        module = None
        integration = name.split(".")[-1]

        ##Simply put: I don't think it'll matter how you set this up, the __init__ will be imported regardless
        ##Messing with the package name and whatever will just cause issues with relative imports and stuff
        ##So: make different file, like integration.py or something, that provided the hooks for the integration_loader
        ##Or simply make writers use the init to provide the correct hooks *shrug*
        
        if name in sys.modules and cls._reload_imports:
            reload_full_module(name)

        if name in sys.modules and not cls._reload_imports:
            module = sys.modules.get(name,None)
        else:
            spec = importlib.util.find_spec(name)

            ##Got this code from: https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported
            if spec  == None:
                _LOGGER.error(f"Unable to import integration {integration} from {name}")
                return
            try:
                module = importlib.util.module_from_spec(spec)
                module = importlib.import_module(name)
            except Exception as exce:
                msg = f"Error importing integration {integration}: {exce}"
                _LOGGER.exception(msg, stack_info=True)
                raise exce

        if not hasattr(module,"async_setup") and not hasattr(module,"setup"):
            _LOGGER.error(f"Integration {integration} is missing the required setup function")
            return

        return module

    @classmethod
    async def async_setup_integrations(cls, core: "CORE", progress_func=None, value_range=()) -> MappingProxyType[Literal["integration_entry"],Any]:
        integration_objects = {}
        
        config = core.config
        if not cls._imported_modules:
            return {}

        if progress_func:
            progress_func(value_range[0] + 1, f"Importing {len(cls._imported_modules)} integrations")
            step = int((value_range[1] - value_range[0] - 1)/len(cls._imported_modules))
            progress = value_range[0] + 1

        for integration, module in cls._imported_modules.items():
            module : "sys.ModuleType"
            setup_func = None
            if hasattr(module,"async_setup"):
                setup_func = module.async_setup
            elif hasattr(module,"setup"):
                setup_func = module.setup

            if not isinstance(setup_func,Callable):
                _LOGGER.error(f"{integration} does not have a valid setup function, not importing")
                raise TypeError
            
            _LOGGER.info(f"Setting up integration {integration}")
            if progress_func:
                progress = progress + step
                progress_func(progress, f"Setting up integration {integration}")
            try:
                if asyncio.iscoroutinefunction(setup_func):
                    res = await setup_func(core, config)
                else:
                    res = setup_func(core, config)

                if res == None:
                    _LOGGER.error(f"Integration setup functions must return a result (at minimum a boolean `True`), or `False`. {integration} returned `None`")
                    continue

                if res == False:
                    _LOGGER.error(f"Something went wrong setting up {integration}")
                    continue
                
                cls._loaded_integrations.add(Path(module.__file__).parent)
                if res != True:
                    integration_objects[integration] = res

                _LOGGER.debug(f"{integration} succesfully set up")
            except (ModuleNotFoundError, ImportError) as exce:
                msg = f"Error importing integration {integration}: {exce}. The integration may not be able to run in an emulated environment."
                _LOGGER.warning(msg)
            except Exception as exce:
                msg = f"Error importing integration {integration}: {exce}."
                _LOGGER.error(msg, exc_info=True)

        return integration_objects
    
    @classmethod
    async def async_start_integrations(cls, core: "CORE"):
        """
        Calls the setup functions for relevant integrations. If a maximum time was set in the config, will continue setting up in the background while inkBoard starts printing.

        Parameters
        ----------
        core : CORE
            The inkBoard core module
        """   

        cls._done_setups = {}
        cls._pending_setups = {}

        config = core.config

        coro_list = set()
        for integration, module in cls._imported_modules.items():
            pkg = module.__package__
            setup_res = core.integration_objects.get(integration,None)

            if hasattr(module, "async_start"):
                if not asyncio.iscoroutinefunction(module.async_start):
                    _LOGGER.error(f"integration {integration}: async_start must be a coroutine")
                    continue
                t = asyncio.create_task(module.async_start(core, setup_res), name=pkg)

            elif hasattr(module, "start"):
                coro = asyncio.to_thread(module.start, core, setup_res)
                t = asyncio.create_task(coro,name=pkg)
            else:
                continue

            coro_list.add(t)
            cls._pending_setups[pkg] = t

        if not coro_list:
            return
        
        setup_time = config.inkBoard["integration_start_time"]

        if setup_time < 0:
            await cls._wait_for_start(coro_list)
        else:
            setup_task = asyncio.create_task(cls._wait_for_start(coro_list))
            try:
                await asyncio.wait_for(asyncio.shield(setup_task), setup_time)
            except asyncio.TimeoutError:
                _LOGGER.warning(f"Integration setup time of {setup_time} seconds elapsed but not all integrations are setup yet. Integrations {cls._pending_setups.keys()} will continue setup in the background.")

        return
    
    @classmethod
    async def _wait_for_start(cls, coro_list):
        pending = coro_list 
        while pending:
            try:
                done, pending = await asyncio.wait(pending,return_when=asyncio.FIRST_COMPLETED)
            except:
                for t in done:
                    t : asyncio.Future
                    if t.cancelled():
                        continue

                    if t.exception() != None:
                        _LOGGER.error(f"Integrations {t.get_name()} ran into an error while starting up: {t.exception()}")
            else:
                names = set()
                for t in done:
                    names.add(t.get_name())
                _LOGGER.info(f"Integrations {names} were started.")
            finally:
                cls._pending_setups.pop(t.get_name())
                cls._done_setups[t.get_name()] = t

    @classmethod
    async def run_integrations(cls, core: "CORE"):
        """
        Runs any long running tasks defined by integrations.
        Exceptions do not interfere with the print loop.

        Parameters
        ----------
        core : CORE
            The inkBoard core module
        """    
        screen = core.screen

        coro_list = set()

        async def await_setup(pkg, runner : asyncio.Task):
            setup_task = cls._pending_setups.get(pkg,None)
            _LOGGER.warning(f"Waiting to start {pkg} long running task until its setup has finished.")
            if setup_task != None:
                await setup_task
                await asyncio.sleep(0)
            await runner

        for integration, module in cls._imported_modules.items():

            pkg = module.__package__
            setup_res = core.integration_objects.get(integration,None)

            if hasattr(module, "async_run"):
                if not asyncio.iscoroutinefunction(module.async_run):
                    _LOGGER.error(f"integration {pkg}: async_run must be a coroutine")
                    continue

                if pkg in cls._pending_setups:
                    coro = await_setup(pkg, module.async_run(core, setup_res))
                else:
                    coro = module.async_run(core, setup_res)

                t = asyncio.create_task(coro, name=pkg)
                coro_list.add(t)

        if not coro_list:
            return

        pending = coro_list
        ##From what I could find, this is the only way to catch out exceptions when they happen (gather only returns when everything is done, or returns a single result)
        ##While keeping a reference to the original task (which doesn't seem to happen for as_completed I believe) -> that only also just throws an exception

        done = []
        while pending:
            try:
                done, pending = await asyncio.wait(pending,return_when=asyncio.FIRST_COMPLETED)
            except:
                for t in done:
                    t : asyncio.Future
                    if t.cancelled():
                        _LOGGER.warning(f"Integration {t.get_name()} was cancelled while running.")

                    if t.exception() != None:
                        _LOGGER.error(f"Integration {t.get_name()} ran into an error while running: {t.exception()}")

    @classmethod
    def _reset(cls):
        cls._integration_keys: dict[str,str] = {}
        cls._installed_integrations: dict[str,Path] = {}
        cls._integration_modules = {}

        cls._imported_modules: dict[str,str] = {}
        cls._loaded_integrations: list[str] = set()




"""
This module takes care of building a dashboard from a config layout.
Its functions can also be called for custom dashboards.
"""

import logging
from typing import TYPE_CHECKING

from ..configuration.loaders import const as yaml_const
from ..helpers import DashboardError

from .loader import DashboardLoader
from .validate import validator_dict

from PythonScreenStackManager.elements import Layout, TabPages, StatusBar

if TYPE_CHECKING:
    from inkBoard import config, core as CORE

logger = logging.getLogger(__name__)

def build_config_elements(config : "config", core: "CORE"):
    """
    Builds all the elements defined in the configuration.
    Order of parsing: elements, layouts, popups, tabs.
    """
    ##Statusbar can simply be parsed from the dict (though should be done here ofc)
    ##main_element or smth needs to be defined somewhere. I think under inkBoard. -> either use a main_tabs key, or simply tabs and take that as default
    ##Default should be None.

    conf = dict(config.configuration)

    DashboardLoader._CORE = core

    if "home_assistant" in conf:
        _add_ha_defaults()

    dash_conf = {}
    for conf_key in yaml_const.DASHBOARD_KEYS:
        if conf_key in conf:
            ##Determine how to deal with the validators, i.e. they should only validate the top_level elements
            validator = validator_dict.get(conf_key,validate.validate_general)
            DashboardLoader._validator = validator

            conf_res = DashboardLoader().construct_dashboard_node(conf[conf_key],True)

            if conf_key  == "statusbar":
                if conf_res == None:
                    conf_res = {}
                else:
                    conf_res = dict(conf_res)
                status_conf = {"size": conf_res.pop("size","?*0.05"),
                                                "location": conf_res.pop("location","top")}
                if "orientation" not in conf_res:
                    if status_conf["location"] in {"top","bottom"}:
                        conf_res["orientation"] = "horizontal"
                    else:
                        conf_res["orientation"] = "vertical"
                try:
                    statusbar = StatusBar(**conf_res)
                    status_conf["element"] = statusbar
                    conf_res = status_conf
                except TypeError as e:
                    logger.error(f"Error in the config for {conf_key}: {e}. Check if the initial arguments were defined correctly.")
                    DashboardLoader._config_error = True
                    continue
            elif conf_key == "main_tabs":
                try:
                    conf_res = TabPages(**conf_res)
                except TypeError as e:
                    logger.error(f"Error in the config for {conf_key}: {e}. Check if the initial arguments were defined correctly.")
                    DashboardLoader._config_error = True
                    continue
                
            dash_conf[conf_key] = conf_res

    if DashboardLoader._config_error:
        raise DashboardError("Error parsing configuration for the dashboard. See the logs for more information.") #@IgnoreExceptions
    return dash_conf

def get_main_layout(dash_config: "config.configuration"):
    """
    Builds the main layout element based on the settings in the configuration.
    Returns the main layout element. Unless otherwise specified in the config, this will return the tab element made from the tabs key, and the statusbar, if specified
    """

    layout = []

    if dash_config.get("main_tabs",None):
        main_elt = dash_config["main_tabs"]

    if "statusbar" in dash_config:
        statusbar_conf = dash_config["statusbar"]
        statusbar = statusbar_conf["element"]
        loc = statusbar_conf["location"]
        if loc in {"left","right"}:
            status_tuple = (statusbar,statusbar_conf["size"])
            if loc == "left":
                statusbar_row = ["?", status_tuple, (main_elt,"?")]
            else:
                statusbar_row = ["?", (main_elt,"?"), status_tuple]
            layout.append(statusbar_row)
        elif loc in {"top", "bottom"}:
            statusbar_row = [statusbar_conf["size"], (statusbar,"?")]
            main_row = ["?", (main_elt,"?")]
            if loc == "top":
                layout.extend([statusbar_row,main_row])
            else:
                layout.extend([main_row,statusbar_row])
    else:
        layout.append(["?", (main_elt,"?")])

    return Layout(layout)

def _add_ha_defaults():
    from inkBoard.core import integration_loader
    imported_modules = integration_loader.imported_integrations
    ha_integration = "homeassistant_client"
    if ha_integration  not in imported_modules:
        return
    
    import importlib

    ha_module = integration_loader._imported_modules[ha_integration]    
    ha_parser = importlib.import_module(".parser",ha_module.__package__)

    ha_elements = getattr(ha_parser,"element_dict",{})
    from . import loader
    loader.default_elements.update(ha_elements)

    


"""
Handles command line arguments for inkboard
"""
import argparse
from . import constants as const
from .logging import LOG_LEVELS

DESIGNER_MOD = const.DESIGNER_INSTALLED

if const.DESIGNER_INSTALLED:
    ##For this: ensure the designer can be imported without the window being build etc.
    ##i.e. move a lot of the init into a runner file
    import inkBoarddesigner
    designer_version = inkBoarddesigner.__version__

def pop_base_args(args) -> dict:
    "Returns a dict with the base argparse arguments removed (i.e. anything that is not command)"
    d = vars(args).copy()
    d.pop("logs")
    d.pop("quiet")
    d.pop("verbose")
    d.pop("command")
    return d


def command_version(*args):
    print(f"inkBoard Version: {const.__version__}")
    if DESIGNER_MOD:
        print(f"inkBoard designer Version: {designer_version}")
    return 0

def command_designer(args):
    if not DESIGNER_MOD:
        print("Running inkBoard designer require the inkBoard designer to be installed")
        print("Run 'pip install inkBoarddesigner' to install it")
        return 1

    inkBoarddesigner.run_designer(args)

def command_install(args):
    from .packaging import command_install
    return command_install(**pop_base_args(args))

def command_pack(args):
    from .packaging import create_config_package
    return create_config_package(**pop_base_args(args))


PRE_CORE_ACTIONS = {
    const.COMMAND_VERSION: command_version,
    const.COMMAND_DESIGNER: command_designer,
    const.COMMAND_INSTALL: command_install
}
"Action that can/have to be run before creating the CORE object"

POST_CORE_ACTIONS = {
    const.COMMAND_PACK: command_pack
}
"Actions to run after creating the CORE object, but before doing any setup otherwise."



def parse_args():

    ##Code layout mainly used from esphome command line interface

    base_parser = argparse.ArgumentParser(add_help=False)
    base_parser.add_argument('--logs',default=None,
                        choices=LOG_LEVELS, help='set log level manually, takes precedent over the --quiet and --verbose flags. If None are set, it defaults to WARNING')    
    base_parser.add_argument('-q', '--quiet', action='store_true', dest='quiet',
                        help="Disables all inkBoard logs")
    base_parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help="Enables all inkBoard logs")

    parser = argparse.ArgumentParser(parents=[base_parser],
                            description="""
                                Aims to make it easy to design dashboards for a variety of devices.
                            """)

    subparsers = parser.add_subparsers(
        help="The mode or command to run inkBoard with:", dest="command", metavar="command"
    )
    subparsers.required = True

    subparsers.add_parser(const.COMMAND_VERSION, 
                        help="Print the inkBoard version and optionally designer version, then exits.")

    parser_run = subparsers.add_parser(
        const.COMMAND_RUN, help="Runs inkBoard using the provided config file"
    )

    parser_run.add_argument(const.ARGUMENT_CONFIG, help="The YAML file used for the dashboard", default=const.DEFAULT_CONFIG)
    ##Could optionally add the RAISE flag to this.

    if DESIGNER_MOD:
        inkBoarddesigner._add_parser(subparsers, const.COMMAND_DESIGNER)
    else:
        designer_parser = subparsers.add_parser(const.COMMAND_DESIGNER, 
                                            description="inkBoard designer is not installed",
                                            help="Runs inkBoard in designer mode. inkBoarddesigner must be installed for it.")

    parser_pack = subparsers.add_parser(
        const.COMMAND_PACK, help="Creates an inkBoard package using the provided config file"
    )

    parser_pack.add_argument(const.ARGUMENT_CONFIG, help="The YAML file used for the package")
    parser_pack.add_argument("name", help="""
                            The name of the package file. Must have no suffix, or end in .zip
                            """, default=None, nargs='?')
    
    parser_pack.add_argument("--all", action='store_true', dest='pack_all', help="Creates a complete package of the config")
    parser_pack.add_argument("--config", action='store_true', help="Includes the configuration file and requires files and folders in the package")
    parser_pack.add_argument("--platform", action='store_true', help="Includes the platform module in the package")
    parser_pack.add_argument("--integrations", action='store_true', help="Includes the loaded integrations in the package")

    parser_install = subparsers.add_parser(
        const.COMMAND_INSTALL, help="Installs inkBoard packages or requirements from a config folder"
    )

    parser_install.add_argument("file", help="""
                            The file to install. If it is a ZIP file, inkBoard will check if it is an inkBoard compatible one (Either an inkBoard package, or a zip of a platform or integration folder).
                            If a YAML file, inkBoard will go through the base directory, the files folder and the custom folder and call pip install on any files it finds titled requirements.txt. In the custom integration folder, it will also take care of installing requirements for the integrations presents.
                            If it is platform or integration, appended by a name, inkBoard will go through the install process of that respective platform/integration, provided it is installed internally.
                            If not supplied, the installer will look for all suitable ZIP files in the current directory.
                            """, default=None, nargs='?')
    ##How to deal with passing the name part of the command?
    
    parser_install.add_argument("name", help="The name of platform or integration to install. This is not used if the file command is not one of integration or platform.",
                                default=None, nargs="?")

    parser_install.add_argument('--no-input', help="Disables any input prompts that are deemed optional", action='store_true')

    parser.add_argument(
        "--version",
        action="version",
        version=f"Version: {const.__version__}",
        help="Prints the inkBoard version and exit.",
    )

    return parser.parse_args()


args = parse_args()



"Holds the typehints for the inkBoard config, mainly typedicts and dataclasses"

from types import MappingProxyType
from typing import *
from pathlib import Path
from dataclasses import dataclass

from inkBoard.helpers import add_required_keys
from inkBoard.logging import LOG_LEVELS

from .const import INKBOARD_FOLDER

if TYPE_CHECKING:
    from PythonScreenStackManager import pssm_types as pssm
    from PythonScreenStackManager.elements import constants as elt
    from mdi_pil import mdiType

##These need to stay as typeddicts since some, like device, need additional options to be set

LogLevels = Union[int, Literal[LOG_LEVELS]]

class _BaseConfigEntry:

    def __getitem__(self, item):
        if isinstance(item,str) and hasattr(self,item): #@IgnoreExceptions
            return getattr(self,item) #@IgnoreExceptions
        else:
            raise KeyError(f"{self.__class__} has no entry {item}") #@IgnoreExceptions

@dataclass(frozen=True)
class InkboardEntry(_BaseConfigEntry):
    """
    Map for inkBoard specific settings. 
    """
    
    name : str
    "The name to give the instance. Used for i.e. the name of the Device in Home Assistant"

    date_format : str = "%-d-%m-%Y"
    """
    Default format to use when representing dates. Default may be device/region specific.
    This site can help giving your the correct format string from a date: https://www.dateformatgenerator.com/?lang=Python&q=17-09-2013 
    """

    time_format : str = "%H:%M"
    """
    Default format to use when representing time. Default may be device/region specific.
    This site can help giving your the correct format string from a time: https://www.dateformatgenerator.com/?lang=Python&q=11%3A11
    """

    integrations : Union[Literal["all"],tuple] = ()
    "List with integrations to load, if they do not have a config key defined (i.e. for custom elements)"

    integration_start_time : 'pssm.DurationType' = -1
    "Maximum time to allow integrations to run their start tasks before inkBoard continues to printing. Gets parsed to seconds. Keep in mind that setting this may cause issues with integrations. Default is -1 (No maximum time.)"

class StylesEntry(TypedDict):
    """
    Dataclass for styling options, used to set custom colours, define default fonts and certain icon styles.
    TypedDict, so can hold any entries, to allow integrations to also define styles.
    """

    menu_header_color: 'pssm.ColorType'
    "Default color for the header bar of menu popups. Defaults to steelblue"

    menu_button_color: 'pssm.ColorType'
    "Default color for menu buttons. Defaults to gray."

    foreground_color: 'pssm.ColorType'
    "Default color for foreground parts of elements, like text colors or icon colors."

    background_color: 'pssm.ColorType'
    "Default color for element backgrounds, when applicable. Also used for calculating automatic contrasting colors when applicable."

    font: str
    "Default font to use for displaying text."

    font_bold: str
    "Default font to use for bold text"

    font_clock: str
    "Default font to use for digital clocks"

    font_size: 'pssm.PSSMdimension'
    "Default size to use for fonts"

    default_icon: 'mdiType'
    "Default icon to use for icon elements"

    missing_icon: 'mdiType'
    "Icon to use when a specified icon or image file cannot be found."

    battery_style: 'elt.batteryIconMapping'
    """
    Default styling for battery icons. Options for the 3 states + default values. Options should hopefully be somewhere in the docs, otherwise see elements.constants.py
    """

    network_style : Literal["lines","signal"]
    "Default style to use for network icons."

    shorthand_colors : dict[str,tuple[int,int,int,int]]
    """
    Allows you to define additional color shorthands.
    Define them as a set of RGBA values, so 4 integers ranging from 0-255
    """

    blur_popup_background: bool
    "Default setting for blurring the background behind popups. Defaults to True"

@dataclass(frozen=True)
class FolderEntry(_BaseConfigEntry):
    """
    Dataclass with the folders inkBoard will look in for custom files. 
    NOT defined in the config, as all folders should have constant names. However, this is used to set the correct Paths when running configs from different folders.
    Eventually it will be made possible however to define these in a way to use custom folders from elsewhere, such that when developing a new dashboard, old files do not need to be changed immediately. However, this option will only be available for the emulator most likely.
    """
    #No need for custom folders, inkBoard automatically makes the same folders in the config folder.

    base_folder : Path
    "Base path leading to the folder containing the config file."

    icon_folder: Path
    "Path pointing to the folder containing user icons"

    picture_folder: Path
    "Path pointing to the folder containing user pictures"

    font_folder: Path
    "Path pointing to the folder user fonts"
    
    custom_folder: Path
    "Path pointing to the folder where users can define custom elements and functions, to be used with the yaml parser."

    screenshot_folder: Path
    "Path pointing to where inkBoard will save screenshots"

    file_folder: Path
    "Path point to the files folder. This folder can be used to hold any files needed that are not yaml configurations. This entire directory will be copied during packing."

@dataclass(frozen=True)
class DesignerEntry(_BaseConfigEntry):
    """
    Dict with settings for the designer, when developing new dashboards. 
    Can also parse certain settings from the device key, like platform, if the key is present.
    """

    height: int = None
    "Optional height to give the screen. Overwrites device settings."

    width: int = None
    "Optional width to give the screen. Overwrites device settings."

    icons: MappingProxyType = MappingProxyType({})
    "Optionally allows one to redefine the icons used for pssm elements in the element tree view,  or set icons for their custom elements."

    platform_validation: bool = True
    "Validates whether the device configuration is valid for the base platform."


class DeviceEntry(TypedDict):
    """
    Dict to configure the device. Mainly important to import the correct device.
    If the emulator key is defined, inkBoard will automatically try and run as emulator using the device settings.
    This config entry accepts any optional arguments to allow for various platforms to be set up via this config.
    """
    
    platform : str
    "The platform (device type) the dashboard will run on, or that will be emulated. A list will be somewhere in the docs."

    model : str
    "Model of the device. May be required for some platforms, others don't need it or can set it automatically."

add_required_keys(DeviceEntry, {"platform"})
deviceMapDefaults = MappingProxyType({"model": None})

@dataclass(frozen=True)
class ScreenEntry(_BaseConfigEntry):
    """
    Map for values having to do with the screen, like interaction or how it is shown.
    """
    
    poll_interval: Union['pssm.DurationType',int, float] = "1min"
    "The amount of time in between polling different attributes, like the Network status."

    close_popup_time: Union[float,'pssm.DurationType', int] = "1min"
    "Default time in which popups will automatically close"

    background: Union[str, 'pssm.ColorType', None] = INKBOARD_FOLDER / "files" / "images" / "default_background.png"
    "Main background of the screen. Can be a color, or an image. If None, the default device background is assumed."

    background_fit: Literal["contain", "cover", "crop", "resize"] = "cover"
    "Method to use to fit the background onto the screen"

    background_fit_arguments: MappingProxyType = MappingProxyType({})
    "Any keyword arguments to pass to the background fit method. Refer to the PIL docs for the arguments for each method"

    backlight_behaviour: Optional[Literal["Manual", "On Interact", "Always"]] = None
    "Behaviour of the device's backlight (if applicable)"

    backlight_time_on: Union['pssm.DurationType',int, float] = "1min"
    "If backlight_behaviour is 'On Interact', this value determines how long the backlight stays on for after the last interaction, by default None"

    touch_debounce_time: 'pssm.DurationType' = '1ms'
    "Time to wait before a touch is considered valid"

    minimum_hold_time: 'pssm.DurationType' = '0.5s'
    "Minimum time required to consider a touch as held, instead of tapped"

    on_interact: Union[Callable[[MappingProxyType, 'pssm.PSSMScreen', 'pssm.CoordType'], None], bool,None] = None
    "Function called whenever a valid touch is registered"

    on_interact_data: MappingProxyType = MappingProxyType({})
    "Any keyword arguments to send along with calls to the on_interact functions"




@dataclass(frozen=True)
class HomeAssistantEntry(TypedDict):
    "Dict with settings required for the home assistant client"

    url : str
    "Url to the home assistant server."

    token: str
    "Long lived access token to authenticate with the server."

    state_colors: MappingProxyType = MappingProxyType({})
    """
    This way you can map the default foreground colors of connected elements to take on the same color when their state matches.
    Also accepts a default entry for unknown states.
    Will be used for elements where state_colors: True is set.
    """

    ping_pong_interval : 'pssm.DurationType' = 30
    "Interval inbetween checking the connection to the server. Generally you can keep this undefined."

    unknown_icon : Optional['mdiType'] = "mdi:help"
    "Default icon to indicate that an entity's state is unknown"

    unavailable_icon : Optional['mdiType'] = "mdi:exclamation-thick"
    "Default icon to indicate that an entity is unavailable"



    ##Also add settings for setting up a socket logger

##Reasons for dict: entries can be omitted
##Namedtuple/dataclass: allows for defaults but requires all entries

@dataclass(frozen=True)
class LoggerEntry(_BaseConfigEntry):
    "Typehint for logging configuration"

    level: LogLevels = "INFO"
    "The base level for logging"

    logs: MappingProxyType[str, LogLevels] = MappingProxyType({})
    "Allows setting log levels for individual components"

    basic_config: Union[MappingProxyType[str,Any],Literal[False],None] = False
    "Forcibly overwrite the basic logging config. This is done before parsing the other logging options, so keep in mind you may overwrite settings."

    log_to_file: Union[MappingProxyType[str,Any],Literal[False],None] = False
    "Settings for logging to a file. This setting is ignored when running in the designer."



class MainEntry(TypedDict):
    "Typehint for the full configuration dict as read out. Not exhaustive"

    inkboard: InkboardEntry
    "Keys for configuring and styling inkBoard"

    device: DeviceEntry
    "Defining what device and platform inkBoard is running on, as well as possible configuring of hardware."

    screen: ScreenEntry
    "Settings that affect how the screen behaves or is interacted with"

    folders: FolderEntry
    "Optionally set paths other than the defaults for folders where custom icons etc. are located"

    styles : StylesEntry
    "Options for styling elements and other things in inkBoard."

    designer : DesignerEntry
    "Settings for the emulator, when designing new dashboards."

    home_assistant : HomeAssistantEntry
    "Settings for the Home Assistant client"

    logger : LoggerEntry
    "Settings to apply for making logs"

__all__ = [
    "InkboardEntry",
    "StylesEntry",
    "FolderEntry",
    "DesignerEntry",
    "DeviceEntry",
    "ScreenEntry",
    "HomeAssistantEntry",
    "LoggerEntry",
    "MainEntry"
]

# rotation : 'pssm.RotationValues' = "UR"
# """
# Screen rotation. Corresponds as follows: \n
# values:\n
#     UR: 'upright' [0°] \n
#     CW: 'clockwise' [90°] \n
#     UD: 'upsidedown' [180°] \n
#     CCW: 'counterclockwise' [270°] \n
# """
##Leaving this in here since it may be useful? But not all devices support it so it should be a device setting.


"Functions to get an inkBoard config set up."

from typing import *
import asyncio
import importlib
from pathlib import Path

import inkBoard
from inkBoard import constants as const, loaders, arguments
from inkBoard.helpers import DeviceError, ScreenError, ConfigError, reload_full_module
import inkBoard.loaders
from inkBoard.logging import setup_logging
import PythonScreenStackManager as PSSM
import inkBoard.platforms

_LOGGER = inkBoard.getLogger(__name__)

if arguments.DESIGNER_MOD:
    import inkBoarddesigner.integrations

if TYPE_CHECKING:
    import inkBoard
    from PythonScreenStackManager import pssm, elements
    from inkBoard import config, core as CORE, platforms


def import_custom_functions(core: "CORE") -> dict[str,Callable]:
    "Imports the modules from custom functions and constructs the dict holding them."

    ##Move this to loader (where integrationloader will be put too.)
    ##Also check what happens when reloading, without reloading the module
    ##I.e. to test what popping from path_importer_cache does -> nothing I can figure out at least

    return inkBoard.loaders.load_custom_functions(core)

def import_custom_elements(core: "CORE") -> dict[str, "elements.Element"]:

    elts = inkBoard.loaders.load_custom_elements(core)

    def parser(elt_str):
        return elts[elt_str]

    core.add_element_parser("custom", parser)
    return

def setup_base_config(config_file) -> "config":
    "Read out the config and Setup the object holding the inkBoard config"

    from inkBoard.configuration import config

    if isinstance(config_file, config):
        return config_file
    
    try:
        config_obj = config(config_file)
    except Exception as exce:
        raise ConfigError(f"Invalid base configuration: {exce}") from exce

    return config_obj

async def setup_device(core: "CORE") -> "platforms.Device":
    "Import the correct platform module and set up its Device instance. Also validate if the device functions correctly."

    from inkBoard import platforms
    ##Gotta check how this deals with being reloaded, same with config
    config = core.config
    try:
        device = platforms.get_device(config, core)
        await asyncio.sleep(0)  ##Allow startup tasks from devices to optionally run
    except Exception as exce:
        if isinstance(exce, DeviceError):
            raise exce
        raise DeviceError(f"Could not set up device: {exce}") from exce
    
    return device

async def setup_screen(core: "CORE") -> "pssm.PSSMScreen":
    "Setup the screen instance that will be handling the printing"

    config = core.config

    from PythonScreenStackManager import pssm

    try:
        screen = pssm.PSSMScreen(core.device, **vars(config.screen))
        await asyncio.sleep(0)
    except Exception as exce:
        raise ScreenError(f"Could not set up pssm screen: {exce}") from exce
    return screen

def setup_styles(core: "CORE"):
    "Applies styles from the config to the style handler"

    ##Will attach style to screen later
    config = core.config

    import PythonScreenStackManager as pssm
    pssm.pssm.styles.Style.add_color_shorthand(**const.INKBOARD_COLORS)

    pssm.constants.SHORTHAND_ICONS["inkboard"] = const.INKBOARD_ICON

    if config.styles:
        pssm.pssm.styles.Style.add_color_shorthand(**config.styles.get("shorthand_colors",{}))
    
    new_folders = {
                "font_folder": config.folders.font_folder,
                "icon_folder": config.folders.icon_folder,
                "picture_folder": config.folders.picture_folder
                }
    pssm.constants.CUSTOM_FOLDERS.update(new_folders)

def setup_dashboard_config(core: "CORE") -> "elements.Layout":
    "Reads out validates the dashboard nodes from the config, as well as optionally importing the custom dashboard file."

    from inkBoard import dashboard

    config = core.config
    dash_conf = dashboard.build_config_elements(config, core)
    main_layout = dashboard.get_main_layout(dash_conf)

    return main_layout

async def setup_integration_loader(core: "CORE"):
    "Sets up the integration loader attached to the core object."

    loader = core.integration_loader

    obj = await loader.async_setup_integrations(core)
    return obj

async def setup_core(config_file, integration_loader: "loaders.IntegrationLoader" = None) -> "CORE":
    "Sets up the core module for running inkBoard, everything up to starting."
    
    from inkBoard import core as CORE

    assert Path(config_file).exists(), f"{config_file} does not exist"
    p = Path(config_file)
    assert Path(config_file).suffix[1:] in const.CONFIG_FILE_TYPES, f"{config_file} must be a yaml file"

    config_folder = Path(config_file).parent

    if integration_loader != None:
        assert not hasattr(CORE,"integration_loader"), "inkBoard core already has an integration loader defined"
        CORE.integration_loader = integration_loader
    
    if CORE.integration_loader:
        folders = {
            "custom.integrations" : config_folder / "custom" / "integrations",
            inkBoard.integrations.__package__: Path(inkBoard.integrations.__path__[0]) 
            }
        if arguments.DESIGNER_MOD:
            folders[inkBoarddesigner.integrations.__package__] = Path(inkBoarddesigner.integrations.__path__[0])
        CORE.integration_loader.get_integrations(folders)
    
    CORE.config = setup_base_config(config_file)

    setup_logging(CORE)

    if CORE.integration_loader:
        CORE.integration_loader.import_integrations(CORE)

    CORE.custom_functions = import_custom_functions(CORE)
    import_custom_elements(CORE)

    CORE.device = await setup_device(CORE)

    CORE.screen = await setup_screen(CORE)
    CORE.screen.add_shorthand_function_group("custom", CORE.parse_custom_function)

    setup_styles(CORE)

    if CORE.integration_loader:
        CORE.integration_objects = await CORE.integration_loader.async_setup_integrations(CORE)

    main_layout = setup_dashboard_config(CORE)

    await CORE.screen.async_add_element(main_layout, skipPrint=True)

    return CORE

async def start_core(core: "CORE"):
    "Starts the inkBoard core, but does not run the print loop yet"

    core.screen.start_batch_writing()

    if core.integration_loader:
        await core.integration_loader.async_start_integrations(core)

async def run_core(core: "CORE"):
    "Runs the inkBoard core. Generally call after `setup_core` and `start_core`"

    coros = [core.screen.async_start_screen_printing(), core.screen._eStop]

    if core.integration_loader:
        coros.append(core.integration_loader.run_integrations(core))
    
    L = asyncio.gather(
                    *coros,
                    return_exceptions=False)
    return await L  #@IgnoreExceptions

def _shutdown_core(core: "CORE", reload_ = False):
    "Shuts down the core object and optionally reloads the necessary modules"
    
    _LOGGER.info("Shutting down inkBoard core")

    for task in asyncio.all_tasks(core.screen.mainLoop):
        if task == asyncio.current_task(core.screen.mainLoop):
            continue
        task.cancel()

def reload_core(core: "CORE", full_reload: bool = False):
    """Reloads the core object and required modules so it can be set up fresh again.

    Parameters
    ----------
    core : CORE
        The inkBoard core
    full_reload : bool, optional
        A full reload reloads all modules that affect printing, i.e. PSSM, non custom integrations, platforms, etc. by default False
    """    

    _shutdown_core(core,full_reload)

    if not full_reload:
        PSSM.pssm._reset()
        for mod in const.BASE_RELOAD_MODULES:
            _LOGGER.debug(f"Reloading module {mod}")
            reload_full_module(mod)
        
        ##This should also take into account modules coming from designer
        ##i.e. insert the specific platform module too if not in inkBoard

    else:
        PSSM.pssm._reset()
        reload_mods = list(const.FULL_RELOAD_MODULES)

        if inkBoard.platforms.__name__ not in core.device.__module__:
            idx = reload_mods.index(inkBoard.platforms.__name__)
            reload_mods.insert(idx+1, core.device.__module__)

            ##Reload integrations here? ##Yeah, can be done via the loader list.

        for mod in reload_mods:
            _LOGGER.debug(f"Reloading module {mod}")
            reload_full_module(mod)

        core.integration_loader._reload_imports = True

    del inkBoard.core
    return


def stop_core(core: "CORE"):
    try:
        _shutdown_core(core,False)
    except Exception as exce:
        print(f"inkBoard did not shutdown gracefully: {exce}")
        return 1
    
    return 0


"Core modules and objects for running inkBoard instances"

from typing import TYPE_CHECKING, Literal, Final, Optional, Any, Callable
from types import MappingProxyType
import logging
from pathlib import Path
from datetime import datetime as dt

import inkBoard
from inkBoard import constants as const
from inkBoard.arguments import args

from . import util  ##Depending on how stuff moves around, may need to import util somewhere else?

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from PythonScreenStackManager.pssm.screen import PSSMScreen
    from PythonScreenStackManager.elements import Element

    from inkBoard.platforms import BaseDevice
    from inkBoard.configuration.configure import config as configuration

    from inkBoard.loaders import IntegrationLoader

IMPORT_TIME = dt.now().strftime("%H:%M:%S")

config: "configuration"
"The config object from the currently loaded configuration"

screen: "PSSMScreen"
"The screen instance managing the screen stack"

device: "BaseDevice"
"The device object managing bindings to the device"

integration_loader: "IntegrationLoader"
"Object responsible for loading integrations"

integration_objects: Final[MappingProxyType[Literal["integration_entry"],Any]]
"Objects returned by integrations"

custom_functions: MappingProxyType[str,Callable]
"Functions in the custom/functions folder of the config"


DESIGNER_RUN : bool = args.command == const.COMMAND_DESIGNER

_INTEGRATION_KEYS = {}
_INTEGRATION_OBJECTS = {}
_ELEMENT_PARSERS = {}

def add_integration_config_key(key : str, folder : Path):
    """
    Adds the key that is connected to an integration and the import function. Will call the import_func if the key is present in the config.

    Parameters
    ----------
    key : str
        The key that users can use in their config file to load in the integration
    folder : Callable
        The function that handles the processing of data under said key.
    """    
    if key in _INTEGRATION_KEYS:
        int_mod = _INTEGRATION_KEYS[key].__module__
        logger.error(f"{key} is already used for a the config of a different integration: {int_mod}")
    else:
        _INTEGRATION_KEYS[key] = folder

def get_integration_config_keys() -> MappingProxyType[str,Callable]:
    return MappingProxyType(_INTEGRATION_KEYS)

def add_element_parser(identifier : str, parser : Callable[[str],"Element"]):
    """
    Adds a parser function for custom elements. The identifier should be unique, and cannot be custom.

    Parameters
    ----------
    identifier : str
        An identifier string that indicates the element is parsed via this parser.
        Does not need the ':'
    parser : Callable[[str],&quot;Element&quot;]
        The function that parses the element. Should return the class, not an instance of the element.
    """    
    if identifier in _ELEMENT_PARSERS:
        logger.error(f"Element identifier {identifier} is already registered")
        return
    
    _ELEMENT_PARSERS[identifier] = parser

def get_element_parsers() -> MappingProxyType:
    "Returns the element parser functions and their identifiers as keys."
    return MappingProxyType(_ELEMENT_PARSERS)

_CUSTOM_FUNC_IDENTIFIER = "custom:"

def parse_custom_function(name: str, attr: str, options = {}) -> Optional[Callable]:
    """Parses a string to a function from the custom functions package. 
    """

    parse_string = name.lower()
    return custom_functions[parse_string]






from datetime import datetime
import sys
from types import MappingProxyType
from typing import Union, TYPE_CHECKING, overload
from pathlib import Path
from datetime import datetime as dt
from contextlib import suppress
import logging

import yaml

from .types import MainEntry

from ..helpers import ConfigError

from . import const
from .types import *
from .loaders import MainConfigLoader, BaseSafeLoader


if TYPE_CHECKING:
    from PythonScreenStackManager import pssm_types as pssm
    from PythonScreenStackManager.elements import constants as elt
    from mdi_pil import mdiType

dashboard_nodes = {}

_LOGGER = logging.getLogger(__name__)

def mount_config_dir(folder: Path):
    
    if not isinstance(folder, str):
        folder = str(folder)

    sys.path.insert(0, folder)
    with suppress(ImportError):
        import custom  # pylint: disable=import-outside-toplevel  # noqa: F401
        _LOGGER.debug("imported custom folder")

    sys.path.remove(folder)
    sys.path_importer_cache.pop(folder, None)

def read_config(filepath: Path) -> MainEntry:
    with open(filepath) as f:
        _config = yaml.load(f, Loader=MainConfigLoader)
    return MappingProxyType(_config)

def set_folders(full_config : MainEntry, base_folder: Path) -> FolderEntry:
    folderDict = {}

    folder_config = full_config.get("folder", {})

    folderDict["base_folder"] = base_folder

    folders = [k for k in FolderEntry.__dataclass_fields__ if "folder" in k and k != "base_folder"]

    ##If folders key is defined, not existing folders should NOT be made (if they are defined)
    for folder in folders:
            if folder in folder_config:
                p = Path(folder_config[folder])
                if not p.is_absolute():
                    p = base_folder / p
            else:

                f = folder.split('_')[0]
                if f != "custom": f = f + 's' 
                p = base_folder / f
                ##Probably do something here to make the folder if it does not exist? Or at least perform a check
                if not Path(p).exists():
                    _LOGGER.warning(f"folder {folder} does not yet exist, making it")
                    pa = Path(p)
                    pa.mkdir()

            folderDict[folder] = p
    

    return FolderEntry(**folderDict)

class configMeta(type):
    """
    Metaclass for inkBoard configuration. Ensures class attributes cannot be overwritten.
    """

    def __setattr__(cls, name, value):
        raise AttributeError("Cannot modify config attributes")

    def __delattr__(cls, name):
        raise AttributeError("Cannot delete config attributes")


##Turn this into a singleton.
##And make it all instance based not class based.
class config(metaclass=configMeta):
    """
    the inkBoard configuration. Automatically build upon important from the configuration file supplied as command line argument.'
    """
    __configFile = const.DEFAULT_CONFIG_FILE
    "The yaml file where the config is read from"

    
    "The full configuration dict as defined in the yaml file."

    __integrations = {}

    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            return object.__new__(cls)
        else:
            raise AttributeError("Can only define one config class. Access it via `inkBoard.core.config`")

    def __init__(self, file: Union[str,Path]):
        if isinstance(file,str):
            file_path = Path(file).resolve()
        else:
            file_path = file.resolve()
        
        
        if not file_path.exists(): raise ConfigError(f"Cannot find file {file}")
        if not file_path.suffix.endswith(const.CONFIG_FILE_TYPES): raise ConfigError(f"Config file must be of type {const.CONFIG_FILE_TYPES}")

        self.__configFile = file
        self.__filePath = file_path
        self.__baseFolder = file_path.parent

        mount_config_dir(self.__baseFolder)
        

        self._readout_time = datetime.now()
        self._readout_time_string = self._readout_time.strftime("%X")

        BaseSafeLoader._base_folder = self.__baseFolder
        BaseSafeLoader.read_secrets()
        BaseSafeLoader.read_entities()

        __full_config : MainEntry = read_config(file_path)

        missing = []
        for entry in const.REQUIRED_KEYS:
            if entry not in __full_config:
                missing.append(entry)
        
        if missing:
            raise ConfigError(f"Missing required keys {missing}")

        self.__full_config = __full_config

        ##Maybe do a few more assertions or some validation to check if the required keys are present?

        self.__folders = set_folders(__full_config, self.__baseFolder)

        ##Apply test in config to see if required keys are present.

        exces = []

        try:
            self.__inkBoard = InkboardEntry(**__full_config["inkBoard"])
        except TypeError as exce:
            exces.append("inkBoard")

        ##Would probably want to somehow validate the device and screen entries beforehand?
        self.__device = MappingProxyType(DeviceEntry(**__full_config["device"]))
        self.__screen = ScreenEntry(**__full_config.get("screen",{}))
        self.__styles = StylesEntry(**__full_config.get("styles",{}))
        self.__logger = LoggerEntry(**__full_config.get("logger",{}))



        try:
            designer : DesignerEntry = DesignerEntry(**__full_config.get("designer",{}))
        except TypeError as exce:
            _LOGGER.error(f"Error in entry designer: {exce}")
            exces.append("designer")

        self.__designer = designer

        if exces:
            raise ConfigError(f"Errors under entries {exces}")

        return

    def __getitem__(self, item: str):
        return self.configuration[item]
    
    @overload
    def get(self, key, /):
        ...
    
    @overload
    def get(self, key, default, /):
        ...

    def get(self, *args):
        "Gets a key from the full config, similar to calling `.get` on a dict"
        if len(args) == 1:
            return self.__full_config[args[0]]
        else:
            return self.__full_config.get(*args)

    @property
    def file(self) -> str:
        "The file used for the configuration. Either a string or the Path instance to it"
        return self.__configFile

    @property
    def included_yamls(self) -> set[str]:
        "All yaml files included in the config, i.e. secrets.yaml and other include constructors."
        return BaseSafeLoader._opened_files

    @property
    def filePath(self) -> Path:
        "Path to the config file. Resolved be absolute."
        return self.__filePath
    
    @property
    def baseFolder(self) -> Path:
        "The folder the configuration file is in"
        return self.__baseFolder

    @property
    def configuration(self) -> MainEntry:
        "The mappingproxy dict that was build by reading out the configuration file."
        return self.__full_config

    @property
    def folders(self) -> FolderEntry:
        "Folder paths for some of the custom folders, useful when running configs from other folders (like the examples)"
        return self.__folders

    @property
    def logger(self) -> LoggerEntry:
        "Settings for the logger"
        return self.__logger

    @property
    def inkBoard(self) -> InkboardEntry:
        "Configuration under the inkboard: entry. Holds settings for the screen instance mainly."
        return self.__inkBoard
    
    @property
    def device(self) -> DeviceEntry: 
        "Settings for the device. Device type/platform and other settings pertaining to hardware not (directly) related to the screen."
        return self.__device

    @property
    def screen(self) -> ScreenEntry:
        "Settings for the screen instance, like refresh time and rotation."
        return self.__screen

    @property
    def styles(self) -> StylesEntry:
        "Styling for inkBoard, For example to overwrite the default font, or define custom color shorthands."
        return self.__styles

    @property
    def designer(self) -> DesignerEntry:
        "Settings for the designer. Empty if not defined"
        return self.__designer
    ##Replace this by using the command line argument


    ##Full config is read out anyways. So add a function for clients to add stuff, which is passed the full config.
    ##Maybe at print start check if every key is read?
    ##Or redo this, such that it has a function to read out everything at a later point, after the init, which can only be called once.



"""
Small library with helper functions for inkBoard, that could not be included in the pssm tools.
Does not import anything from pssm or integration yet, and is mainly meant to supplement some small useful functions.
"""
import typing   ##Need to import typing to evaluate union strings (those are converted to typing.Union[...])
from typing import TYPE_CHECKING
from typing import TypedDict, Generic, TypeVar, Callable, Union, Literal
from types import ModuleType
import functools
import logging 
import sys
import importlib
from contextlib import suppress

if TYPE_CHECKING:
    from PythonScreenStackManager.elements import Element

_LOGGER = logging.getLogger("inkBoard")

ph = TypedDict("ph")
"Placeholder typedict for typehinting"

class InkBoardError(Exception):
    "Base Exception for inkBoard"

class DeviceError(InkBoardError):
    "Something went wrong setting up the device"

class ScreenError(InkBoardError):
    "Something went wrong setting up the screen instance"

class ConfigError(InkBoardError):
    "Something is wrong with the configuration"

class DashboardError(ConfigError):
    "Unable to setup the dashboard"

class QuitInkboard(InkBoardError):
    "Exception to set as eStop to quit the current inkBoard session"
    pass

def add_required_keys(td : ph, keys : frozenset):
    """
    Adds the required keys to the typeddict, and removes them from the optional keys

    Parameters
    ----------
    td : TypedDict
        The typed dict to add the keys to
    keys : frozenset
        the keys to add like {"key1","key2"}
    """
    td.__required_keys__ = td.__required_keys__.union(keys)
    td.__optional_keys__ = td.__optional_keys__.difference(td.__required_keys__)

def add_optional_keys(td : ph, keys : frozenset):
    """
    Adds the optional keys to the typeddict, and removes them from the required keys

    Parameters
    ----------
    td : TypedDict
        The typed dict to add the keys to
    keys : frozenset
        the keys to add like {"key1","key2"}
    """
    td.__optional_keys__ = td.__optional_keys__.union(keys)
    td.__required_keys__ = td.__required_keys__.difference(td.__required_keys__)

def check_required_keys(typeddict : ph, checkdict : dict, log_start : str):
    """
    checks if the keys required by typedict are present in checkdict. Exits inkboard if any are missing.
    Uses name for constructing logs.

    Parameters
    ----------
    typeddict : TypedDict
        The TypedDict to get required keys from
    checkdict : dict
        The dict to check
    log_start : str
        Starting string for log messages; {log_start} is missing ...
    """

    missing = {}
    for k in typeddict.__required_keys__:
        if k not in checkdict:
            missing.add(k)
        if missing:
            _LOGGER.error(f"{log_start} is missing required {'entries' if len(k) > 1 else 'entry'} {k}, exiting inkBoard.")
            sys.exit()
    return missing

##May move some things of these to a util module
def reload_full_module(module: Union[str,ModuleType], exclude: list[str] = []):    
    """Reloads the module and all it's submodules presently imported

    Keep in mind reloading imports the module, so things can behave unexpectedly, especially if order matters.
    Generally be careful using this, reloading does not mean imports in none reloaded modules are refreshed, which can cause issuess when i.e. calling `isinstance`
    
    Parameters
    ----------
    module : Union[str,ModuleType]
        The base module to reload.
    exclude : list[str]
        List with module names that do not get excluded. Names need to match in full.
    """    
    if isinstance(module, ModuleType):
        module = module.__package__
    
    if isinstance(exclude,str):
        exclude = [exclude]

    mod_list = [x for x in sys.modules.copy() if x.startswith(module) and not x in exclude]
    for mod_name in mod_list:
        mod = sys.modules[mod_name]
        try:
            importlib.reload(mod)
        except ModuleNotFoundError:
            _LOGGER.error(f"Could not reload module {mod_name}")
    
    for mod_name in mod_list:
        sys.modules.pop(mod_name)
    
    return

def function_parameter_dict(func: Callable, types_as_str: bool = False, is_method: bool = False) -> dict[Literal["required", "optional"], dict[str, dict[Literal["type_hint","default"],str]]]:
    """_summary_

    Parameters
    ----------
    func : Callable
        The function to create the parameter dict of
    types_as_str : bool
        If True, Any type hints will be converted into a string representation. This means either using the types __name__ if present, otherwise just casting it to a string.
    is_method : bool
        If True, the function is considered a method, and the first argument (generally self or cls) will be omitted

    Returns
    ----------
    dict:
        A dict with the keys required and optional, for required parameters and optional parameters.
        If a parameter has a type hint, it is included as a string in the dict for said parameter under 'type_hint'. For optional parameters, their default values are included as well.
    """
    if func.__defaults__:
        num_defaults = len(func.__defaults__)
        default_values = func.__defaults__
    else:
        num_defaults = 0
        default_values = []

    f_code = func.__code__
    num_required = f_code.co_argcount - num_defaults

    func_vars = f_code.co_varnames[:f_code.co_argcount]

    if is_method:
        req_args = func_vars[1:num_required]
    else:
        req_args = func_vars[:num_required]

    opt_args = func_vars[num_required:]

    type_hints = func.__annotations__


    required = {}
    for var_name in req_args:
        if var_name in type_hints:
            hint = type_hints[var_name]
            if types_as_str:
                hint = getattr(hint,"__name__") if hasattr(hint, "__name__") else str(hint)
            required[var_name] = {"type_hint": hint}
        else:
            required[var_name] = {}

    optional = {}
    for i, var_name in enumerate(opt_args):
        optional[var_name] = {"default": default_values[i]}

        if var_name in type_hints:
            hint = type_hints[var_name]
            if types_as_str:
                if types_as_str:
                    hint = getattr(hint,"__name__") if hasattr(hint, "__name__") else str(hint)
            optional[var_name].update({"type_hint": hint})

    return {"required": required, "optional": optional}


T = TypeVar('T')

class classproperty(Generic[T]):
    "Used to avoid the deprecation warning (and the extra writing) needed to set class properties"
    
    def __init__(self, method: Callable[..., T]):
        self.method = method
        functools.update_wrapper(self, wrapped=method) # type: ignore

    def __get__(self, obj, cls=None) -> T:
        if cls is None:
            cls = type(obj)
        return self.method(cls)


"""
Platforms for pssm devices.
"""
from pathlib import Path
import logging
import importlib
import importlib.util
from typing import TYPE_CHECKING
import sys

from .. import constants as const

from .basedevice import BaseDevice, Device, FEATURES
from .validate import validate_device

if const.DESIGNER_INSTALLED:
    import inkBoarddesigner

if TYPE_CHECKING:
    from inkBoard import config as configuration, core as CORE

_LOGGER = logging.getLogger(__name__)

def get_device(config : "configuration", core: "CORE") -> Device:
    "Initialises the correct device based on the config."

    ##Don't forget to include a way to import the designer
    if core.DESIGNER_RUN:
        from inkBoarddesigner.emulator.device import Device, window
        return Device(config)

    platform_package = __package__
    conf_platform = config.device["platform"]
    platform_name = conf_platform
    if const.DESIGNER_INSTALLED:
        platform_path = Path(inkBoarddesigner.__file__).parent / "platforms" / conf_platform
        platform_package = f"{inkBoarddesigner.__package__}.platforms"
        if not platform_path.exists() or not platform_path.is_dir():
            platform_path = Path(__file__).parent / conf_platform
            platform_package = __package__
    elif not "/" in conf_platform:
        platform_path = Path(__file__).parent / conf_platform
    else:
        platform_path = Path(conf_platform)
        if conf_platform.startswith("./"):
            platform_path = config.baseFolder / platform_path
        _LOGGER.info(f"Looking for custom device at {platform_path}")
        platform_name = platform_path.name
        platform_package = f"{platform_package}.{platform_name}"
    

    if not platform_path.exists() or not platform_path.is_dir():
        _LOGGER.error(f"Device platform {conf_platform} does not exist.")
        raise ModuleNotFoundError(f"Device platform {conf_platform} does not exist.")
    else:
        ##Test this again once the platform is installed
        dev_spec = importlib.util.spec_from_file_location(platform_package, str(platform_path / "__init__.py"), submodule_search_locations=[])
        device_platform: basedevice = importlib.util.module_from_spec(dev_spec)
        sys.modules[platform_package] = device_platform
        dev_spec.loader.exec_module(device_platform)
        device_platform = importlib.import_module(".device",platform_package)

    device_args = dict(config.device)
    device_args.pop("platform")
    device = device_platform.Device(**device_args) #-> pass the config to this right -> no but the device mappingproxy
    validate_device(device)
    return device


"Constants"

import __main__
from pathlib import Path

from ..constants import INKBOARD_FOLDER, CONFIG_FILE_TYPES

DEFAULT_CONFIG_FILE = "configuration.yaml"
BASE_FOLDER = Path.cwd()

SECRETS_YAML = 'secrets.yaml'
ENTITIES_YAML  = 'entities.yaml'

REQUIRED_KEYS = (
    "inkBoard",
    "device",
)

DASHBOARD_KEYS = (
    "elements",
    "layouts",
    "popups",
    "main_tabs",
    "statusbar"
    )
"Config keys that indicate they are used to parse the dashboard, which means they will not be processed passed the first node when reading out the config initially."


"Logging classes for inkBoard."

import logging
import logging.handlers
from pathlib import Path
from typing import Any, Optional, TYPE_CHECKING
from functools import partial, partialmethod
from contextlib import suppress
from dataclasses import asdict
from types import MappingProxyType

try:
    # Python 3.7 and newer, fast reentrant implementation
    # without task tracking (not needed for that when logging)
    from queue import SimpleQueue as Queue
except ImportError:
    from queue import Queue

if TYPE_CHECKING:
    from inkBoard import core as CORE
    from inkBoard.configuration.types import LoggerEntry


NOTSET = logging.NOTSET
VERBOSE = int(logging.DEBUG/2)
DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
WARN = WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL
FATAL = logging.FATAL

LOG_LEVELS = ("NOTSET", "VERBOSE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")



log_format = '%(asctime)s [%(levelname)s %(name)s %(funcName)s, line %(lineno)s]: %(message)s'
log_dateformat = '%d-%m-%Y %H:%M:%S'


class ANSICOLORS:
    GRAY = "\x1b[38;5;247m"
    GREEN = "\x1b[32;20m"
    BLUE =  "\x1b[34;20m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m"
    BOLD_RED = "\x1b[31;1m"
    RESET =  "\x1b[0m"

ANSI_FORMATS = {
        logging.NOTSET: ANSICOLORS.GRAY,
        logging.DEBUG: ANSICOLORS.GREEN,
        logging.INFO: ANSICOLORS.BLUE,
        logging.WARNING: ANSICOLORS.YELLOW,
        logging.ERROR: ANSICOLORS.RED,
        logging.CRITICAL: ANSICOLORS.BOLD_RED
    }


class LogFormats:
    fmt = log_format
    datefmt = log_dateformat


_LOGGER = logging.getLogger(__name__)


class BaseLogger(logging.Logger):
    "Logger class with the verbose function defined for type hinting purposes"

    def __init__(self, name, level = 0):
        super().__init__(name, level)

    def verbose(self, msg, *args, exc_info = None, stack_info = False, stacklevel = 1, extra = None):
        "Logs a message at VERBOSE level (below DEBUG)"
        return self.log(VERBOSE, msg, *args, exc_info = None, stack_info = False, stacklevel = 1, extra = None)



class BaseFormatter(logging.Formatter):
    
    formatter = logging.Formatter(log_format, log_dateformat)

    @classmethod
    def format(cls, record):
        return cls.formatter.format(record)


class ColorFormatter(logging.Formatter):
    
    def __init__(self, fmt = log_format, datefmt = log_dateformat, style = "%", validate = True):
        super().__init__(fmt, datefmt, style, validate)

    def format(self, record):
        formatted = BaseFormatter.format(record)
        if record.levelno < DEBUG:
            format_level = 0
        elif record.levelno in ANSI_FORMATS:
            format_level = record.levelno
        else:
            for level in ANSI_FORMATS:
                format_level = level
                if level >= record.levelno:
                    break

        prefix = ANSI_FORMATS.get(format_level,"default")
        return f"{prefix}{formatted}{ANSICOLORS.RESET}" 

class InkBoardQueueHandler(logging.handlers.QueueHandler):
    ##Altered from HA logger

    listener: Optional[logging.handlers.QueueListener] = None

    def handle(self, record: logging.LogRecord) -> Any:
        """Conditionally emit the specified logging record.

        Depending on which filters have been added to the handler, push the new
        records onto the backing Queue.

        The default python logger Handler acquires a lock
        in the parent class which we do not need as
        SimpleQueue is already thread safe.

        See https://bugs.python.org/issue24645
        """
        return_value = self.filter(record)
        if return_value:
            self.emit(record)
        return return_value

    def close(self) -> None:
        """Tidy up any resources used by the handler.

        This adds shutdown of the QueueListener
        """
        super().close()
        if not self.listener:
            return
        self.listener.stop()
        self.listener = None

streamhandler = logging.StreamHandler()
streamhandler.setFormatter(ColorFormatter(log_format, log_dateformat))

def init_logging(log_level: str = None, quiet: bool = False, verbose: bool = False) -> None:
    """Initialises the logger, such that the messages printed to stdout are color coded.
    
    Done before setting up queue handler, such that messages printed before reading the config are also logged.
    """

    logging.setLoggerClass(BaseLogger)

    logging.addLevelName(VERBOSE, "VERBOSE")
    logging.Logger.verbose = partialmethod(logging.Logger.log, VERBOSE)
    logging.verbose = partial(logging.log, VERBOSE)

    logging.basicConfig(format=log_format, 
                    datefmt=log_dateformat,
                    handlers=[streamhandler])
    base_logger = logging.getLogger()
    if log_level:
        base_logger.setLevel(log_level)
    elif verbose:
        base_logger.setLevel(VERBOSE)
    elif quiet:
        base_logger.setLevel(CRITICAL)
    else:
        base_logger.setLevel(WARNING)

    ##Would it be better to set up the stream handler already? I'm not quite sure

def overwrite_basicConfig(core: "CORE", config: "LoggerEntry"):
    "Overwrites the basicConfig of logging"

    base_args = asdict(config.basic_config)
    
    logging.basicConfig(**base_args, 
                        handlers=[streamhandler],
                        force=True
                        )
    
    new_format = base_args.get("format", log_format)
    new_datefmt = base_args.get("datefmt", log_dateformat)
    new_style = base_args.get("style", "%")
    new_formatter = logging.Formatter(new_format, new_datefmt, new_style)
    BaseFormatter.formatter = new_formatter


def setup_filehandler(core: "CORE", config: "LoggerEntry"):
    "Sets up the rotating filderhandler logs"

    if isinstance(config.log_to_file, (dict,MappingProxyType)):
        fileconf = config.log_to_file
    else:
        fileconf = {}
    
    fileconf.setdefault("backupCount", 5)
    fileconf.setdefault("filename", "inkboard.log") ##Default filename: logs -> but resolve to config folder.
    if isinstance(fileconf["filename"], str):
        name = fileconf["filename"]
        if "/" in name or "\\" in name:
            filename = Path(name)
        else:
            name = Path(name)
            filename = core.config.baseFolder / "logs" / name
            with suppress(FileExistsError):
                filename.resolve().parent.mkdir()
                _LOGGER.debug(f"Made folder for logs at {filename.parent}")

    else:
        filename = fileconf["filename"]
        assert isinstance(filename,Path),  "logging to file filename must be a string or Path"

    ##Check is performed for non custom too, but those folder will have just been made if not already there
    assert filename.parent.absolute().exists(), "Logging to custom locations requires the folder to exist"

    do_rollover = False
    if filename.exists():
        do_rollover = True

    fileconf["filename"] = filename
    file_handler = logging.handlers.RotatingFileHandler(**fileconf)
    file_handler.setFormatter(BaseFormatter())
    if do_rollover:
        file_handler.doRollover()

    logging.root.addHandler(file_handler)

    return

def setup_logging(core: "CORE"):
    "Sets up logging via the config definitions"
    
    config = core.config.logger

    if config.basic_config != False:
        overwrite_basicConfig(core, config)

    if config.log_to_file != False:
        setup_filehandler(core, config)
    
    ##Remote logging: setup later
    ##Need more knowledge on best practices, as well as knowing what is the best way to set up server/client for general connections

    logging.root.setLevel(config.level)
    for log_name, level in config.logs:
        logging.getLogger(log_name).setLevel(level)

    queue = Queue()
    queue_handler = InkBoardQueueHandler(queue)
    logging.root.addHandler(queue_handler)

    migrated_handlers: list[logging.Handler] = []
    for handler in logging.root.handlers[:]:
        if handler is queue_handler:
            continue
        logging.root.removeHandler(handler)
        migrated_handlers.append(handler)

    listener = logging.handlers.QueueListener(
        queue, *migrated_handlers, respect_handler_level=True)
    queue_handler.listener = listener
    
    listener.handlers

    listener.start()



"""
yaml safeloader classes for the inkBoard config
"""

import sys
import logging
from types import MappingProxyType
from typing import TYPE_CHECKING
from string import Template
from pathlib import Path

import yaml

from yaml import SafeLoader as FastestAvailableSafeLoader
try:
    from yaml import CSafeLoader as FastestAvailableSafeLoader
except ImportError:
    pass
    
from . import const

if TYPE_CHECKING:
    from yaml import SafeLoader as FastestAvailableSafeLoader

_LOGGER = logging.getLogger(__name__)

def secret_constructor(loader: "BaseSafeLoader", node: yaml.nodes.ScalarNode) -> str:
    
    key = loader.construct_scalar(node)
    if key in loader._secrets:
        return loader._secrets[key]
    else:
        _LOGGER.error(f"{key} is not defined in secrets.yaml")

def entity_constructor(loader: "BaseSafeLoader", node: yaml.nodes.ScalarNode) -> str:
    
    ##Probably see if the yaml line can be logged here.
    key = loader.construct_scalar(node)
    if key in loader._entities:
        return loader._entities[key]
    else:
        _LOGGER.error(f"{key} is not defined in entitities.yaml")

def include_constructor(loader: "BaseSafeLoader", node: yaml.nodes.ScalarNode) -> str:
    """Get appropriate entries from secrets.yaml"""
    file = loader.construct_scalar(node)
    file_path = Path(file)
    if not file_path.is_absolute():
        file_path = loader._base_folder / file_path
    
    if not file_path.exists():
        _LOGGER.error(f"Included file {file} cannot be found in {str(loader._base_folder)}")
        return {}

    with open(file_path) as f:
        c = yaml.load(f, Loader=BaseSafeLoader)

    return c

class BaseSafeLoader(FastestAvailableSafeLoader):
    "Base config loader for inkBoard. Used to register tags and the like."
    
    _base_folder: Path = None
    _substitutions: dict
    _opened_files = set()

    def __init__(self, stream):
        self.__class__._opened_files.add(stream.name)
        super().__init__(stream)

    @classmethod
    def read_secrets(cls):
        secrets_file = cls._base_folder / const.SECRETS_YAML
        if secrets_file.exists():
            with open(secrets_file) as f:
                cls._secrets = MappingProxyType(yaml.load(f, Loader=cls))
                return
        else:
            cls._secrets = MappingProxyType({})

    @classmethod
    def read_entities(cls):
        entity_file = cls._base_folder / const.ENTITIES_YAML
        if entity_file.exists():
            with open(entity_file) as f:
                cls._entities = MappingProxyType(yaml.load(f, Loader=cls))
                return
        else:
            cls._entities = MappingProxyType({})

    def construct_scalar(self, node):
        val = super().construct_scalar(node)
        if "$" in val and hasattr(self, "_substitutions"):
            val = Template(val).safe_substitute(**self.__class__._substitutions)
        return val

BaseSafeLoader.add_constructor("!secret",secret_constructor)
BaseSafeLoader.add_constructor("!entity",entity_constructor)
BaseSafeLoader.add_constructor("!include", include_constructor)

class MainConfigLoader(BaseSafeLoader):
    
    def __init__(self, stream):
        self._top_node = True
        super().__init__(stream)

    def construct_mapping(self, node, deep=False):

        d = {}
        if self._top_node:
            self._top_node = False
            parse_later = {}

            for (key_node, value_node) in node.value:
                if key_node.value == "substitutions":
                    if isinstance(value_node, yaml.MappingNode):
                        val = super().construct_mapping(value_node)
                    else:
                        val = self.construct_object(value_node)
                    
                    d[key_node.value] = val
                    BaseSafeLoader._substitutions = MappingProxyType(val)

                elif key_node.value in const.DASHBOARD_KEYS:
                    _LOGGER.verbose(f"dashboard node is {value_node}")
                    d[key_node.value] = value_node
                else:
                    parse_later[key_node.value] = value_node

            for node_name, value_node in parse_later.items():
                    if isinstance(value_node,yaml.ScalarNode):
                        val = self.construct_scalar(value_node)
                    elif isinstance(value_node, yaml.SequenceNode):
                        val = self.construct_sequence(value_node)
                    elif isinstance(value_node, yaml.MappingNode):
                        val = super().construct_mapping(value_node)
                    else:
                        _LOGGER.warning(f"Unknown node type {value_node}")

                    d[node_name] = val
        else:
            d = super().construct_mapping(node, deep)
        #Not returning mappingproxies, as it leads to quite some difficulties
        #I.e. JSON not wanting to dump stuff when it is a MappingProxy.
        return d



"""
A basedevice needed to run inkBoard. Mainly an extension from the pssm basedevice.
"""

from abc import abstractmethod
from typing import Optional
import asyncio

from functools import cached_property
from collections import namedtuple

import inkBoard
from PythonScreenStackManager.pssm_types import *
from PythonScreenStackManager import tools
from PythonScreenStackManager.devices import PSSMdevice, DeviceFeatures, \
                                    Battery as BaseBattery, Backlight as BaseBacklight, Network as BaseNetwork
from PythonScreenStackManager.devices.const import FEATURES, _attr_list, DeviceFeatures
import inkBoard.constants

from ..helpers import function_parameter_dict

_LOGGER = inkBoard.getLogger(__name__)

class FEATURES(FEATURES):

    FEATURE_CONNECTION = inkBoard.constants.FEATURE_CONNECTION
    "Subfeature of network, indicates the device is able to manage the network connection (i.e. disconnect, connect, etc.)"

_attr_list = _attr_list.copy()   ##Copying so the original is not altered
_attr_list.extend([y for x, y in FEATURES.__dict__.items() if not x.startswith("_")])
_DeviceTuple = namedtuple("_DeviceTuple", _attr_list, defaults=(False,)* len(_attr_list))


class InkboardDeviceFeatures(_DeviceTuple, DeviceFeatures, FEATURES):
    """Class for indicating which features the device has. 
    
    Pass all the features using the constants from the `FEATURES` class.
    Do not interface with this class directly, but use `device.has_feature(FEATURES.FEATURE_{})` instead.
    """

    def __new__(cls, *features: str, **kwargs):
        for a in features:
            kwargs[a] = True
        return _DeviceTuple.__new__(cls, **kwargs)

featurehint = Literal[tuple(_attr_list)]

class BaseDevice(PSSMdevice):
    "Base device class for inkBoard"

    ##Need some properties in here, as well as abstractmethods
    ##At least: platform and model
    ##And rewrite the name property?
    
    def __init__(self, features: DeviceFeatures, screenWidth: int, screenHeight: int, viewWidth: int, viewHeight: int, screenMode: str, imgMode: str, defaultColor: ColorType, model: str = None, name: str = None, **kwargs):
        
        self._model = None
        super().__init__(features, screenWidth, screenHeight, viewWidth, viewHeight, screenMode, imgMode, defaultColor, name, **kwargs)

    #region
    @classproperty
    def platform(cls) -> str:
        "The device platform"
        mod = cls.__module__.split('.')
        if mod[-1] == "device":
            return mod[-2]
        return mod[-1]

    
    @cached_property
    def model(self) -> Optional[str]:
        "The model of the device"
        return self._model
    #endregion

    def has_feature(self, feature: Union[str,featurehint]):
        return super().has_feature(feature)

    async def async_pol_features(self):
        """Pols the necessary features of the device to see if their state has changed.

        The screen calls this automatically at preset intervals.
        """

        if self.has_feature(FEATURES.FEATURE_NETWORK):
            await self.network.async_update_network_properties()

        if self.has_feature(FEATURES.FEATURE_BATTERY):
            await self.battery.async_update_battery_state()

        return

    def _set_screen(self):

        super()._set_screen()

        if self.has_feature(FEATURES.FEATURE_POWER):
            self.Screen.add_shorthand_function("power-off", tools.wrap_to_tap_action(self.power_off))
            self.Screen.add_shorthand_function("reboot", tools.wrap_to_tap_action(self.reboot))


class BaseConnectionNetwork(BaseNetwork):
    "Abstract base class to manage a device's network connection."

    @abstractmethod
    async def async_connect(self, ssid: str = None, password: str = None):
        """Base async method to connect to a network."""
        return

    def connect(self, ssid: str = None, password: str = None):
        "Connects to a wifi network"
        asyncio.create_task(self.async_connect(ssid,password))

    @abstractmethod
    async def async_disconnect(self):
        """Base async method to disconnect from the network."""

    def disconnect(self):
        "Disconnects to a wifi network"
        asyncio.create_task(self.async_disconnect())


class Device(BaseDevice):
    """An inkBoard device

    This class is meant as a most basic implementation. Reference the documentation for platforms for the full API of its device.    
    """

    @property
    def network(self) -> Union[BaseNetwork,BaseConnectionNetwork]: return

    @property
    def backlight(self) -> BaseBacklight: return

    @property
    def battery(self) -> BaseBattery: return

def create_emulatorjson_init(device: type[Device]):
    """Updates the platform's emulator.json file with the parameters from the device's __init__ function.

    This is meant as a dev function, and imports a couple of modules too. Generally don't use it aside from when developing platforms.

    Parameters
    ----------
    device : type[Device]
        The _class_ to process. The platform is automatically gathered from this.
    """    

    ##Since this is meant as a developer function that is not ran often, these imports are done in function
    import json
    import inspect
    from pathlib import Path

    if isinstance(device,BaseDevice):
        device = device.__class__

    assert issubclass(device, BaseDevice), "device must a class which is a subclass of the BaseDevice"

    _LOGGER.info(f"Gathering init parameters for device platform {device.platform}")

    parameter_dict = function_parameter_dict(device.__init__, True, True)


    device_folder = Path(inspect.getfile(device)).parent
    platform_file = device_folder / "emulator.json"

    if platform_file.exists():
        _LOGGER.info("Opening emulator.json")
        with open(platform_file) as f:
            cur_conf = json.load(f)
    else:
        cur_conf = {}

    _LOGGER.info("Writing to emulator.json")
    with open(platform_file, "w") as f:
        cur_conf.update({"__init__": parameter_dict})
        json.dump(cur_conf, f, indent=4)

    _LOGGER.info("Succesfully updated emulator.json")
    return

def create_emulatorjson_features(device: Device):
    """Uses an _instance_ of a device to add/update features to the emulator.json file

    Parameters
    ----------
    device : Device
        _description_
    """    

    import json
    import inspect
    from pathlib import Path

    assert isinstance(device._features, InkboardDeviceFeatures), "Use the InkboardDeviceFeatures class for the features attribute"

    _LOGGER.info("Updating device features in emulator.json")

    device_folder = Path(inspect.getfile(device.__class__)).parent
    platform_file = device_folder / "emulator.json"

    if platform_file.exists():
        _LOGGER.info("Opening emulator.json")
        with open(platform_file) as f:
            cur_conf = json.load(f)
    else:
        cur_conf = {}

    _LOGGER.info("Writing to emulator.json")
    with open(platform_file, "w") as f:
        cur_conf.update({"features": device._features._asdict()})
        json.dump(cur_conf, f, indent=4)

    return



from multiprocessing import freeze_support

import inkBoard.logging

freeze_support()

import logging
from typing import Union, TYPE_CHECKING
from pathlib import Path
import asyncio
import sys
import concurrent.futures

import inkBoard
from inkBoard import constants as const, bootstrap, loaders
from PythonScreenStackManager.exceptions import ReloadWarning, FullReloadWarning
from inkBoard.arguments import args, PRE_CORE_ACTIONS, POST_CORE_ACTIONS

if TYPE_CHECKING:
    from inkBoard import core as CORE
    import PythonScreenStackManager
    from inkBoard.configuration.configure import config

_LOGGER = inkBoard.getLogger(__name__)
importer_thread = concurrent.futures.ThreadPoolExecutor(None,const.IMPORTER_THREADPOOL)

async def run_inkBoard(config_file):
    "Runs inkBoard from the passed config file"

    while True:
        CORE = await bootstrap.setup_core(config_file, loaders.IntegrationLoader)
        
        if args.command in POST_CORE_ACTIONS:
            return POST_CORE_ACTIONS[args.command](args)
        
        await bootstrap.start_core(CORE)
        
        try:
            await bootstrap.run_core(CORE)
        except FullReloadWarning:
            await asyncio.sleep(0)
            bootstrap.reload_core(CORE, full_reload=True)
        except ReloadWarning:
            await asyncio.sleep(0)
            bootstrap.reload_core(CORE)
        except (SystemExit, KeyboardInterrupt):
            await asyncio.sleep(0)
            _LOGGER.info("Closing down inkBoard")
            ##Check if it is needed to close
            return bootstrap.stop_core(CORE)
        except Exception as exce:
            _LOGGER.error(f"Something unexpected went wrong running inkBoard: {exce}")
            _LOGGER.warning("Attempting graceful shutdown")
            return bootstrap.stop_core(CORE)
        
        await asyncio.sleep(0)

def run():
    "Starts the main eventloop and runs inkBoard. This function is blocking"
    res = asyncio.run(run_inkBoard(args.configuration),
            debug=_LOGGER.getEffectiveLevel() <= logging.DEBUG)
    return res

def run_config(config_file: Union[Path,str]):
    "Starts the main eventloop and runs inkBoard, using the given config file. This function is blocking"
    return asyncio.run(run_inkBoard(config_file),
                        debug=_LOGGER.getEffectiveLevel() <= logging.DEBUG)

def main():
    inkBoard.logging.init_logging(args.logs, args.quiet, args.verbose)

    if args.command in PRE_CORE_ACTIONS:
        return PRE_CORE_ACTIONS[args.command](args)
    
    ##Maybe make a core folder similar to ESPhome, to hold things like the config file, screen instance etc.
    return run()

if __name__ == "__main__":
    sys.exit(main())



"""
The yaml loader that parses the dashboard config (element types and the like)
"""

import yaml
import logging
from typing import Callable, Literal
from pathlib import Path

from PythonScreenStackManager import elements
from PythonScreenStackManager.pssm.screen import DuplicateElementError

from .. import core as CORE
from ..configuration import loaders, const

from .validate import validate_general

logger = logging.getLogger(__package__)

## Loading here kinda messes with things.
##Reload before building then. -> should be fine since it is not interfaced with

##This should be fine to do here since it imports from PSSM
default_elements = CORE.util.get_module_elements(elements)

#elements that are part of the default pack, i.e. parsed without an identifier

class DashboardLoader(loaders.BaseSafeLoader):
    """
    A special yaml loader that is able to construct elements from dicts with settings.
    The type of element is parsed using the `'type'` keyword, so be mindful when using that (in general, don't do it).
    Integrations can register prefixes for element types (for example ha or custom) which allows users to use elements from those integrations by setting the type to `ha:Climate`, for example.
    It can be used directly on a yaml document/stream, however it's main use is to parse the dashboard nodes that were not parsed into a dict by the main config loader.
    For this, you can instantiate the class without a stream argument, and on that instance call `construct_dashboard_node` to parse the elements defined in that node. The main advantage of this is that it preserves the line numbers in the yaml document, which makes for clearer error messages.
    """
    ##The constructors have an added depth argument, which indicates, somewhat, at what depth the node is generated, since the deep parameter tends to be true, i.e. it functions recursively and hence should work
    ##This is mainly used to set the correct validator, since when validating layouts, there may still be elements defined in the layouts themselves.

    _config_error = False
    _CORE: CORE = None

    _validator : Callable[[Literal["Element_class"],Literal["Requested_type"]],None] = validate_general
    "Function used to validate parsed elements. Passed to the parser function."

    def __init__(self, stream = None):
        if stream != None:
            super().__init__(stream)

    def parse_element_type(self, elt_type: str, validator: Callable[[Literal["Element_class"],Literal["Requested_type"]],None] = validate_general) -> elements.Element:
        
        ##Maybe put the parsing logic entirely in core?
        if elt_type == "None" and validator == validate_general:
            return None
        
        if ":" not in elt_type:
            return default_elements[elt_type]
        else:
            idf, elt_type_str = elt_type.split(":")
            parsers = self._CORE.get_element_parsers()
            if idf not in parsers:
                msg = f"No integration registered the element identifier {idf}"
                logger.error(msg)
                raise SyntaxWarning(msg)
            else:
                parser = parsers[idf]
                elt_class = parser(elt_type_str)

        validator(elt_class,elt_type)

        return elt_class

    def construct_mapping(self, node : yaml.MappingNode, deep=True, depth = 0):

        d = {}
        for (key_node, value_node) in node.value:
            val = self.construct_dashboard_node(value_node,deep, depth=depth+1)
            d[key_node.value] = val

        if "type" not in d:
            return d

        if d["type"] == "None" and len(d) == 1:
            return None

        if depth <= 1:
            validator = DashboardLoader._validator
        else:
            validator = validate_general

        try:
            elt_type = self.parse_element_type(d["type"], validator)
        except (TypeError, KeyError):
            yaml_line = node.start_mark.line
            if "id" in d:
                msg = f"Invalid element type '{d['type']}' (id {d['id']}) in configuration file {Path(node.start_mark.name).name}, line {yaml_line}"
            else:
                msg = f"Invalid element type '{d['type']}' in configuration file {Path(node.start_mark.name).name}, line {yaml_line}"
            logger.error(msg)
            self.__class__._config_error = True
            return None
        except SyntaxWarning:
            yaml_line = node.start_mark.line
            if "id" in d:
                msg = f"Invalid element identifier in configuration file {Path(node.start_mark.name).name}, line {yaml_line}: {d['type']}  (id {d['id']})"
            else:
                msg = f"Invalid element identifier in configuration file {Path(node.start_mark.name).name}, line {yaml_line}: {d['type']}"
            logger.error(msg)
            self.__class__._config_error = True
            return None
        
        if elt_type == None:
            return d
        
        type_str = d.pop("type")
        
        try:
            elt = elt_type(**d)
        except TypeError as e:
            yaml_line = node.start_mark.line
            elt_str = type_str
            if "id" in d:
                elt_str = f"[{elt_str}: {d['id']}]"
            msg = f"Error constructing element {type_str} in configuration file {node.start_mark.name}, line {yaml_line}: {e}"
            logger.error(msg)
            self.__class__._config_error = True
            return None
        except DuplicateElementError as e:
            yaml_line = node.start_mark.line
            if "id" in d:
                elt_id = d["id"]
                msg = f"An element with id {elt_id} has already been registered. Duplicate element is located in configuration file {node.start_mark.name}, line {yaml_line}."
            else:
                msg = f"Element {type_str} in configuration file {node.start_mark.name}, line {yaml_line} got a duplicate ID: {e}"
            logger.error(msg)
            self.__class__._config_error = True
            return None
        return elt

    def construct_sequence(self, node, deep = True, depth = 0):
        seq_vals = []
        for sequence_node in node.value:
            v = self.construct_dashboard_node(sequence_node, deep, depth=depth+1)
            seq_vals.append(v)
        return seq_vals
        
    def construct_dashboard_node(self, node, deep, depth = 0):
        if isinstance(node, yaml.MappingNode):
            v = self.construct_mapping(node,deep, depth)
        elif isinstance(node, yaml.SequenceNode):
            v = self.construct_sequence(node,deep, depth)
        elif isinstance(node, yaml.ScalarNode):
            if getattr(node,"tag",None) in self.yaml_constructors:
                tag_constructor = self.yaml_constructors[node.tag]
                v = tag_constructor(self,node)
            else:
                v = self.construct_scalar(node)
        return v






import inspect
from types import ModuleType
from pathlib import Path

from PythonScreenStackManager.elements import Element

##In core since it needs to be reloaded when pssm reloads
def get_module_elements(module: ModuleType) -> dict[str,"Element"]:
    """
    Creates a dict with all the valid elements in a module to use in element parsers, for example

    Parameters
    ----------
    module : ModuleType
        The module to inspect. It is asserted to be a python module

    Returns
    -------
    dict[str,`Element`]
        A dict with the names users can use as type in the yaml config, and the actual class is represents
    """    

    ##See if this can be rewritten to ensure it does not need to be imported everytime.
    ##Probably just add an import for Element in there.
    # from PythonScreenStackManager.elements import Element #Just to prevent any preliminary imports, don't import it globally
    assert inspect.ismodule(module), "Module must be a module type"

    element_dict = {}
    for name, cls in inspect.getmembers(module, inspect.isclass):
        if (module.__name__ in cls.__module__
            and issubclass(cls,Element) 
            and not inspect.isabstract(cls) 
            and name[0] != "_"):
            element_dict[name] = cls

    return element_dict




"""
Some functions that validate whether the parsed element type is of the correct type for the section it is being defined in.
These functions raise type errors if they are not, otherwise they simply return.
"""

import logging
from inspect import isabstract

from PythonScreenStackManager import elements

logger = logging.getLogger(__package__)

def validate_general(elt_class, requested_type: str):
    """
    Validates if the provided class is a subclass of the PSSM Element class. `requested_type` is a string for the log messages.
    Also tests whether the class still has abstractmethods, or its name starts with an '_', in which case it is also deemed invalid.
    """
    
    if not issubclass(elt_class,elements.Element):
        msg = f"Requested element type: {requested_type} did not return an Element type, returned {elt_class}"
        logger.error(msg)
        raise TypeError(msg)
    elif isabstract(elt_class) or elt_class.__name__.startswith("_"):
        msg = f"Requested element type: {requested_type} returned {elt_class}, which is a building block element and cannot be used."
        logger.error(msg)
        raise TypeError(msg)

def validate_layout(elt_class, requested_type: str):
    """
    Validates if the provided class is a subclass of the PSSM Layout class
    Keep in mind that, due to way elements have been set up, various elements like Tiles and Counters also evaluate to a layout.
    However, layouts can also simply be defined under element's, and the distinction is moreso to reduce clutter.
    """
    
    validate_general(elt_class,requested_type)

    if not issubclass(elt_class,elements.Layout):
        msg = f"Requested element type: {requested_type} did not return a Layout type Element, returned {elt_class}"
        logger.error(msg)
        raise TypeError(msg)
    
def validate_popup(elt_class, requested_type: str):
    "Validates if the provided class is a subclass of the PSSM Popup class"
    
    validate_general(elt_class, requested_type)
    
    if not issubclass(elt_class,elements.Popup):
        msg = f"Requested element type: {requested_type} did not return a Layout type Element, returned {elt_class}"
        logger.error(msg)
        raise TypeError(msg)
    
validator_dict = {
    "elements": validate_general,
    "layouts": validate_layout,
    "popups": validate_popup,
}
"Dict mapping config entries to possible validator functions"


"Handles inkBoard packages, both creating and installing them."

import asyncio
import zipfile
import os
import tempfile
import shutil
import inspect
import json
import subprocess
import sys

from typing import TYPE_CHECKING, TypedDict, Literal, Callable, Union, Optional
from abc import abstractmethod
from functools import partial
from pathlib import Path
from datetime import datetime as dt
from contextlib import suppress

import inkBoard
import inkBoard.platforms
from inkBoard.configuration.const import CONFIG_FILE_TYPES, INKBOARD_FOLDER
from inkBoard.types  import *
from inkBoard import constants as const, bootstrap

import PythonScreenStackManager as PSSM


if TYPE_CHECKING:
    from inkBoard import core as CORE
    from packaging.version import Version

with suppress(ModuleNotFoundError):
    import inkBoarddesigner

try:
    from packaging.version import parse as parse_version
except ModuleNotFoundError:
    from pkg_resources import parse_version



_LOGGER = inkBoard.getLogger(__name__)

packagetypes = Literal['package', 'integration', 'platform']
packageidfiles : dict[packagetypes,str] = {
    'package': 'package.json',
    'integration': 'manifest.json',
    'platform': 'platform.json'
}
internalinstalltypes = Literal["platform", "integration"]

class PackageDict(TypedDict):

    created: str
    "The date and time the package was created, in isoformat"

    created_with: Literal["inkBoard", "inkBoarddesigner"]
    "Whether this package was created via inkBoard itself, or via the designer"

    versions: dict[Literal["inkBoard", "PythonScreenStackManager", "inkBoarddesigner"],str]
    "The versions of the core packages installed when creating it. Designer version is None if not installed"

    platform: str
    "The platform the package was created for"

class NegativeConfirmation(UserWarning):
    "Raised by ask confirm if the confirmation was negative"

INKBOARD_PACKAGE_INTERNAL_FOLDER = ".inkBoard"
#Folder name where files from a package are put which are gotten from or destined to the site-packages inkBoard folder.

VERSION_COMPARITORS = ('==', '!=', '>=', '<=', '>', '<')
"Comparison operators allowed for versioning, so they can be evaluated internally"

DESIGNER_FILES = {"designer", "designer.py"}

REQUIREMENTS_FILE = 'requirements.txt'

required_attributes = {
    "config",
    "device",
    "integration_loader"
    
}

InkboardVersion = parse_version(inkBoard.__version__)
PSSMVersion = parse_version(PSSM.__version__)

def get_comparitor_string(input_str: str) -> Literal[VERSION_COMPARITORS]:
    "Returns the comparitor (==, >= etc.) in a string, or None if there is None."
    if c := [x for x in VERSION_COMPARITORS if x in input_str]:
        return c[0]
    return

def compare_versions(requirement: Union[str,"Version"], compare_version: Union[str,"Version"]) -> bool:
    """Does simple version comparisons.

    For requirements, accepts both a general requirement string (i.e. '1.0.0'), or a comparison string (i.e. package < 1.0.0))

    Parameters
    ----------
    requirement : Union[str,&quot;Version&quot;]
        The requirement to test
    compare_version : Union[str,Version]
        The version to compare the requirement to

    Returns
    -------
    bool
        True if the requirement is satisfied, false if not
    """

    if isinstance(compare_version,str):
        compare_version = parse_version(compare_version)

    if not isinstance(requirement, str):
        ##To be sure that the pkg_resources Version is also fine
        return compare_version >= requirement

    if c := [x for x in VERSION_COMPARITORS if x in requirement]:
        req_version = requirement.split(c[0])[-1]   ##With how the comparitors are set up, 0 should always be the correct one
        comp_str = f"compare_version {c[0]} required_version"
    else:
        req_version = requirement
        comp_str = f"compare_version >= required_version"
    
    return eval(comp_str, {}, {"compare_version": compare_version, "required_version": parse_version(req_version)})

def confirm_input(msg: str, installer: "BaseInstaller"):
    answer = input(f"{msg}\n(Y/N): ")
    if answer.lower() in {"y","yes"}:
        return True
    elif answer.lower() in {"n","no"}:
        return False
    else:
        print("Please answer one of Y(es) or N(o) (Not case sensitive)")
        return confirm_input(msg)

def create_config_package(configuration: str, name: str = None, pack_all: bool = False, config: bool = False, platform: bool = False, integrations: bool = False):
    """Sets up a core instance and creates a package from it

    Parameters
    ----------
    configuration : str
        The YAML file to use
    name : str, optional
        The name of the package, by default None
    pack_all : bool, optional
        Packages all components (config stuff, platform and integrations), by default False
    config : bool, optional
        Packages the config folder, by default False
    platform : bool, optional
        Packages the platform, by default False
    integrations : bool, optional
        Packages the imported integrations, by default False

    Returns
    -------
    int
        Return code
    """    
    core = asyncio.run(bootstrap.setup_core(configuration, bootstrap.loaders.IntegrationLoader))
    return create_core_package(core, name, pack_all, config, platform, integrations)

def create_core_package(core: "CORE", name: str = None, pack_all: bool = False, config: bool = False, platform: bool = False, integrations: bool = False):
    """Creates an inkBoard package from a core instance.

    This bundles all required files and folders from the configuration folder, as well in the required platforms and integrations.

    Parameters
    ----------
    core : CORE
        The core object constructed from the config
    """

    if pack_all:
        Packager(core).create_package(name)
    else:
        pack = []
        if config: pack.append("configuration")
        if platform: pack.append("platform")
        if integrations: pack.append('integration')

        Packager(core).create_package(name, pack)
    return 0


def command_install(file: str, name: str, no_input: bool):
    ##Add functionality to installer for internal installs (platforms and integrations)
    ##Usage: install [platform/integration] [name]

    if file in internalinstalltypes.__args__:
        return install_internal(file, name, no_input)
    else:
        return install_packages(file, no_input)

def install_internal(install_type: str, name:str, no_input: bool = False):
    return InternalInstaller(install_type, name, no_input, confirm_input).install()

def install_packages(file: Union[str, Path] = None, no_input: bool = False):
    
    if file:
        return PackageInstaller(file, skip_confirmations=no_input, confirmation_function=confirm_input).install()
    else:
        packages = PackageInstaller.gather_inkboard_packages()
        if len(packages) == 1:
            print(f"Found 1 package that can be installed")
        else:
            print(f"Found {len(packages)} packages that can be installed")
        for package in packages:
            ##Add a confirmation message for each file.
            PackageInstaller(package, skip_confirmations=no_input, confirmation_function=confirm_input).install()
        return 0


class Packager:
    """Takes care of creating inkBoard packages from configs
    """

    def __init__(self, core: "CORE", folder: Union[str,Path] = None, progress_func: Callable[[str,str, float],None] = None):
        self.CORE = core
        self.config = core.config
        if folder:
            if isinstance(folder,str): folder = Path(folder)
            assert folder.is_dir(), "Folder must be a directory"
            self.base_folder = folder
        else:
            self.base_folder = core.config.baseFolder
        self._copied_yamls = set()
        self.__progress_func = progress_func
    
    def report_progress(self, stage: str, message: str, progress: float):
        "Reports progress to the progress function, if any"
        if self.__progress_func:
            self.__progress_func(stage, message, progress)
        else:
            _LOGGER.info(message)

    def create_package(self, package_name: str = None, pack: list[Literal['configuration', 'platform', 'integration']] = ['configuration', 'platform', 'integration']):

        self.report_progress("Start", f"Creating a package for {self.CORE.config.file}", 0)

        self.report_progress("Gathering", "Creating temporary directory", 5)

        with tempfile.TemporaryDirectory(dir=self.base_folder) as tempdir:

            if 'configuration' in pack:
                self.report_progress("Configuration", "Copying configuration directory", 10)
                self.copy_config_files(tempdir)
            
            if 'platform' in pack:
                self.report_progress("Platform", "Copying platform directory", 30)
                self.copy_platform_folder(tempdir)

            if 'integration' in pack:
                self.report_progress("Integrations", "Copying included integrations", 50)
                self.copy_integrations(tempdir)

            self.report_progress("Package Info", "Creating Package info file", 70)            

            package_info = self.create_package_dict()
            with open(Path(tempdir) / "package.json", 'w') as f:
                json.dump(package_info, f, indent=4)

            if not package_name:
                package_name = f'inkBoard_package_{package_info["platform"]}_{self.CORE.config.filePath.stem}'

            self.report_progress("Zip File", "Creating Package zipfile", 75)
            _LOGGER.info("Creating package zip file")

            zipname = self.base_folder / f'{package_name}.zip'
            with zipfile.ZipFile(zipname, 'w') as zip_file:
                for foldername, subfolders, filenames in os.walk(tempdir):
                    _LOGGER.verbose(f"Zipping contents of folder {foldername}")
                    for filename in filenames:
                        file_path = os.path.join(foldername, filename)
                        zip_file.write(file_path, os.path.relpath(file_path, tempdir))
                    for dir in subfolders:
                        dir_path = os.path.join(foldername, dir)
                        zip_file.write(dir_path, os.path.relpath(dir_path, tempdir))

            self.report_progress("Done", f"Package created: {zipname}", 100)
            _LOGGER.info(f"Package created: {zipname}")

        return

    def copy_config_files(self, tempdir):
        "Copies all files and folders from the config directory in to the temporary folder"

        _LOGGER.info(f"Copying files from config folder {self.base_folder}")
        config_dir = Path(tempdir) / "configuration"
        config_folders_copy = {
        "icon", "picture", "font", "custom", "file"
        }


        for folder_attr in config_folders_copy:
            ignore_func = partial(self.ignore_files, self.config.folders.custom_folder / "integrations", 
                            ignore_in_baseparent_folder = DESIGNER_FILES)
            path: Path = getattr(self.config.folders, f"{folder_attr}_folder")
            if not path.exists():
                continue
            
            _LOGGER.info(f"Copying config folder {path.name}")
            shutil.copytree(
                src= path,
                dst= config_dir / path.name,
                ignore=ignore_func
            )
        
        for yamlfile in self.config.included_yamls:
            if Path(yamlfile) in self._copied_yamls:
                _LOGGER.debug(f"Yaml file {yamlfile} was already copied.")
                continue

            _LOGGER.debug(f"Copying yaml file {yamlfile}")
            shutil.copy2(
                src=yamlfile,
                dst=config_dir
            )

        _LOGGER.info(f"Succesfully copied contents of config folder.")
        return
    
    def copy_platform_folder(self, tempdir):
        
        tempdir = Path(tempdir)

        if self.CORE.DESIGNER_RUN:
            platform = self.CORE.device.emulated_platform
            platform_folder = self.CORE.device.emulated_platform_folder
        else:
            platform = self.CORE.device.platform
            platform_folder = Path(inspect.getfile(self.CORE.device.__class__)).parent
        
        _LOGGER.info(f"Copying platform {platform} from {platform_folder} to package")

        manual_files = {"readme.md", "install.md", "installation.md", "package_files"}
        manual_dir = (tempdir / "configuration") if (tempdir / "configuration").exists() else tempdir

        for file in platform_folder.iterdir():
            if file.name.lower() not in manual_files:
                continue

            _LOGGER.debug(f"Copying platform manual file {file}")
            if file.is_dir():
                shutil.copytree(
                    src = file,
                    dst = manual_dir / "files",
                    dirs_exist_ok=True
                )
            else:
                manual_files.add(file.name)
                shutil.copy2(
                    src = file,
                    dst = manual_dir
                )

        ignore_func = partial(self.ignore_files, platform_folder.parent, ignore_in_baseparent_folder = manual_files | DESIGNER_FILES )
        _LOGGER.debug("Copying platform folder")
        shutil.copytree(
                src = platform_folder,
                dst = tempdir / INKBOARD_PACKAGE_INTERNAL_FOLDER / "platforms" / platform_folder.name,
                ignore = ignore_func
            )
        
        _LOGGER.info("Succesfully copied platform folder")
        return

    def copy_integrations(self, tempdir):
        
        tempdir = Path(tempdir)

        ##Filter out integrations from the custom folder
        all_integrations = self.CORE.integration_loader.imported_integrations

        _LOGGER.info("Copying all non custom integrations to package")
        for integration, location in all_integrations.items():
            if location.is_relative_to(self.config.folders.custom_folder):
                ##Skip integrations here. Those were already copied during the config folder phase
                continue
            _LOGGER.debug(f"Copying integration {integration}")
            ignore_func = partial(self.ignore_files, location.parent, ignore_in_baseparent_folder=DESIGNER_FILES)
            shutil.copytree(
                src= location,
                dst= tempdir / INKBOARD_PACKAGE_INTERNAL_FOLDER / "integrations" / location.name,
                ignore=ignore_func
            )
            
        _LOGGER.info("Succesfully copied integrations")
        return

    def ignore_files(self, parentbase_folder: Path, src, names, ignore_in_baseparent_folder: set = {}):
        """Returns a list with files to not copy for `shutil.copytree`

        Parameters
        ----------
        parentbase_folder : Path
            The base folder being copied from
        src : str
            source path, passed by `copytree`
        names : list[str]
            list with file and folder names, passed by `copytree`
        ignore_in_baseparent_folder : set, optional
            Set with filenames to ignore (i.e. not copy), _Only if_ the parent folder of `src` is `base_ignore_folder`, by default {}

        Returns
        -------
        _type_
            _description_
        """        

        ignore_set = {"__pycache__"}
        if Path(src).parent == parentbase_folder:
            ignore_set.update(ignore_in_baseparent_folder)

        for name in filter(lambda x: x.endswith(CONFIG_FILE_TYPES), names):
            self._copied_yamls.add(Path(src) / name)

        return ignore_set

    def create_package_dict(self) -> PackageDict:
        
        package_dict = {"created": dt.now().isoformat()}

        package_dict["versions"] = {"inkBoard": inkBoard.__version__,
                    "PythonScreenStackManager": PSSM.__version__}
        

        if self.CORE.DESIGNER_RUN:
            package_dict["created_with"] = "inkBoarddesigner"
            package_dict["versions"]["inkBoarddesigner"] = inkBoarddesigner.__version__
            package_dict["platform"] = self.CORE.device.emulated_platform
        else:
            package_dict["created_with"] = "inkBoard"
            package_dict["versions"]["inkBoarddesigner"] = None
            package_dict["platform"] = self.CORE.device.platform

        return PackageDict(**package_dict)

class BaseInstaller:
    """Base class for installers

    Call `Installer().install()` to run the installer, or use the pip functions to install packages via pip
    """

    _skip_confirmations: bool
    _confirmation_function: Callable[[str, 'BaseInstaller'],bool]

    @property
    def skip_confirmations(self) -> bool:
        "Whether to ask for confirmation for all actions"
        return self._skip_confirmations
    
    @property
    def confirmation_function(self) -> Callable[[str, 'BaseInstaller'],bool]:
        "The function used to prompt the user for confirmation"
        return self._confirmation_function
    

    @abstractmethod
    def install(self):
        "Runs the installer"
        return

    def install_platform_requirements(self, name: str, platform_conf: platformjson) -> bool:
        """Installs requirements based on a platformjson dict

        Parameters
        ----------
        name : str
            Name of the platform. For logging
        platform_conf : platformjson
            The platform.json dict

        Returns
        -------
        bool
            Whether the requirements were installed successfully
        """        

        platform = name
        requirements = platform_conf["requirements"]
        if requirements:
            res = self.pip_install_packages(*requirements, no_input=self._skip_confirmations)

            if res.returncode != 0:
                try:
                    msg = f"Something went wrong installing the requirements using pip. Continue installation of platform {platform}?"
                    self.ask_confirm(msg, force_ask=True)
                except NegativeConfirmation:
                    return False

        for opt_req, reqs in platform_conf.get("optional_requirements", {}).items():
            with suppress(NegativeConfirmation):
                msg = f"Install requirements for optional features {opt_req}?"
                self.ask_confirm(msg)
                self.pip_install_packages(*reqs, no_input=self._skip_confirmations)
        
        return True

    def install_integration_requirements(self, name: str, manifest: manifestjson) -> bool:
        """Installs integration requirements based on a manifestjson dict

        Parameters
        ----------
        name : str
            Name of the integration. For logging
        platform_conf : platformjson
            The manifest.json dict

        Returns
        -------
        bool
            Whether the requirements were installed successfully
        """

        integration_version = parse_version(manifest['version'])
        integration = name

        _LOGGER.info(f"Installing new Integration {integration}, version {integration_version}")
        
        requirements = manifest["requirements"]
        if requirements:
            res = self.pip_install_packages(*requirements, no_input=self._skip_confirmations)

            if res.returncode != 0:
                try:
                    msg = f"Something went wrong installing the requirements using pip. Continue installation of integration {integration}?"
                    self.ask_confirm(msg, force_ask=True)
                except NegativeConfirmation:
                    return False

        for opt_req, reqs in manifest.get("optional_requirements", {}).items():
            with suppress(NegativeConfirmation):
                msg = f"Install requirements for optional features {opt_req}?"
                self.ask_confirm(msg)
                self.pip_install_packages(*reqs, no_input=self._skip_confirmations)
        return True

    def ask_confirm(self, msg: str, force_ask: bool = False):
        """Prompts the user to confirm something.

        Calls the confirmation function passed at initialising if skip_confirmations is `False`
        
        Parameters
        ----------
        msg : str
            The message to pass to the confirmation function
        force_ask : bool
            Force the prompt to appear, regardless of the value passed to `skip_confirmations`

        Raises
        ------
        NegativeConfirmation
            Raised if the confirmation does not evaluate as `True`
        """
        if self._skip_confirmations and not force_ask:
            return
        
        if self._confirmation_function:
            if not self._confirmation_function(msg, self):
                raise NegativeConfirmation
        return

    def check_inkboard_requirements(self, ib_requirements: inkboardrequirements, required_for: str) -> bool:
        """Checks if inkBoard requirements are met for the current install

        Performs version checks for inkBoard and pssm, and checks for installed platforms and their versions. 

        Parameters
        ----------
        ib_requirements : inkboardrequirements
            Dict with inkBoard specific requirements
        required_for : str
            What the requirements are required for. Used in log messages. Best practice is to pass it as [type] [name], e.g. 'Platform desktop'

        Returns
        -------
        bool
            `True` if requirements are met, otherwise `False`.
        """        

        ##Check: required inkboard version, pssm version and required integrations/platforms 
        warn = False
        if v := ib_requirements.get("inkboard_version", None):
            if not compare_versions(v, InkboardVersion):
                warn = True
                _LOGGER.warning(f"{required_for} requirment for inkBoard's version not met: {v}")

        if v := ib_requirements.get("pssm_version", None):  ##I think this should generally be met by having the inkBoard requirement met tho?
            if not compare_versions(v, PSSMVersion):
                warn = True
                _LOGGER.warning(f"{required_for} requirment for PSSM's version not met: {v}")

        for platform in ib_requirements.get('platforms', []):
            req_vers = None
            if c := get_comparitor_string(platform):
                platform, req_vers = platform.split(c)
            
            if not (INKBOARD_FOLDER / "platforms" / platform).exists():
                warn = True
                _LOGGER.warning(f"Platform {platform} required  for {required_for} is not installed")
                ##Should maybe check this in regards with package installing? i.e. if these are otherwise present in the package
                ##But will come later.
            elif req_vers:
                with open(INKBOARD_FOLDER / "platforms" / platform / packageidfiles["platform"]) as f:
                    platform_conf: platformjson = json.load(f)
                    cur_version = platform_conf["version"]

                if not compare_versions(c + req_vers, cur_version):
                    warn = True
                    _LOGGER.warning(f"Platform {platform} does not meet the version requirement: {c + req_vers}")

        ##And do the same for integrations.
        for integration in ib_requirements.get('integrations', []):
            req_vers = None
            if c := get_comparitor_string(integration):
                integration, req_vers = integration.split(c)
            
            if not (INKBOARD_FOLDER / "integrations" / integration).exists():
                warn = True
                _LOGGER.warning(f"Integration {integration} required for {required_for} is not installed")
                ##Should maybe check this in regards with package installing? i.e. if these are otherwise present in the package
                ##But will come later.
            elif req_vers:
                with open(INKBOARD_FOLDER / "integrations" / integration / packageidfiles["integration"]) as f:
                    integration_conf: manifestjson = json.load(f)
                    cur_version = integration_conf["version"]

                if not compare_versions(c + req_vers, cur_version):
                    warn = True
                    _LOGGER.warning(f"Integration {integration} does not meet the version requirement: {c + req_vers}")

        return not warn

    @staticmethod
    def pip_install_packages(*packages: str, no_input: bool = False) -> subprocess.CompletedProcess:
        """Calls the pip command to install the provided packages

        Parameters
        ----------
        packages : str
            The packages to install (as would be passed to pip as arguments)
        no_input: bool
            Disables prompts from pip        

        Returns
        -------
        subprocess.CompletedProcess
            The result of the subprocess.run function
        """

        if not packages:
            return

        if no_input:
            args = [sys.executable, '-m', 'pip', '--no-input', 'install', *packages]
        else:
            args = [sys.executable, '-m', 'pip', 'install', *packages]

        res = subprocess.run(args)
        return res
    
    @staticmethod
    def pip_install_requirements_file(file: Union[str,Path], *, no_input: bool = False) -> subprocess.CompletedProcess:
        """Calls the pip command to install the provided .txt file with requirements

        Parameters
        ----------
        file : Union[str,Path]
            The text file holding the requirements
        no_input: bool
            Disables prompts from pip
        
        Returns
        -------
        subprocess.CompletedProcess
            The result of the subprocess.run function
        """

        if isinstance(file,Path):
            file = str(file.resolve())

        if no_input:
            args = [sys.executable, '-m', 'pip', '--no-input', 'install', '-r', file]
        else:
            args = [sys.executable, '-m', 'pip', 'install', '-r', file]


        res = subprocess.run(args)
        return res

    ##Options to install:
    # - Package
    # - Platform
    # - Integration
    # - requirements; internal and external -> internal eh, should be taken care of when actually installing it.

class PackageInstaller(BaseInstaller):
    """Installs an inkBoard compatible .zip file, or requirements files in a config directory.

    Call `PackageInstaller().install()` to run the installer.
    There are a few classmethods and staticmethods too that can be called without instantiating.

    Parameters
    ----------
    file : Union[Path,str]
        The .zip file to install
    skip_confirmations : bool, optional
        Skips most confirmation messages during installation, except those deemed vital, by default False
    confirmation_function : Callable[[str, Installer],bool], optional
        Function to call when asking for confirmation, gets passed the question to confirm and the Installer instance., by default None
    """

    def __init__(self, file: Union[Path,str], skip_confirmations: bool = False, confirmation_function: Callable[[str, 'BaseInstaller'],bool] = None):
        self._file = Path(file)
        assert self._file.exists(), f"{file} does not exist"
        self._confirmation_function = confirmation_function
        self._skip_confirmations = skip_confirmations

        if self._file.suffix in CONFIG_FILE_TYPES:
            self._package_type = "configuration"
        else:
            self._package_type: packagetypes = self.identify_zip_file(self._file)
        return

    def install(self):
        """Runs the appropriate installer for the package type.
        """

        if self._package_type == "integration":
            self.install_integration()
        elif self._package_type == "platform":
            self.install_platform()
        elif self._package_type == "package":
            self.install_package()
        elif self._package_type == "configuration":
            self.install_config_requirements(self._file, self._skip_confirmations, self._confirmation_function)

    def install_package(self) -> Optional[packagetypes]:
        """Installs a package type .zip file file
        """        

        file = self._file

        if self._package_type != 'package':
            raise TypeError(f"{file} is not a package type .zip file")

        try:
            self.ask_confirm(f"Install package {file}?")
        except NegativeConfirmation:
            _LOGGER.info(f"Not installing package {file}")
            return

        with zipfile.ZipFile(file) as zip_file:
            self.__zip_file = zip_file
            zip_path = zipfile.Path(zip_file)
            # with zip_file.open(packageidfiles["package"]) as f:
                ##This section is used to determine compatibility of the package and the installed modules
            f = zip_file.open(packageidfiles["package"])
            package_info: PackageDict = json.load(f)

            vers_msg = ""
            if (v := parse_version(package_info["versions"]["inkBoard"])) >= InkboardVersion:
                vers_msg = vers_msg + f"Package was made with a newer version of inkBoard ({v}). Installed is {InkboardVersion}."
            
            if (v := parse_version(package_info["versions"]["PythonScreenStackManager"])) >= PSSMVersion:
                vers_msg = vers_msg + f"Package was made on with a newer version of PSSM ({v}). Installed is {PSSMVersion}."
            
            if vers_msg:
                print(vers_msg)
                try:
                    self.ask_confirm(f"Version mismatch, continue installing {self._package_type}?")
                except NegativeConfirmation:
                    return
            
            ##Check if platform is installed or present in the package.
            package_platform = package_info["platform"]

            if ((INKBOARD_FOLDER / "platforms" / package_platform).exists() or 
                (zip_path / INKBOARD_PACKAGE_INTERNAL_FOLDER / "platforms" / package_platform).exists()):
                pass
            else:
                msg = f"Package was made for platform {package_platform}, but it is not installed or present in the package. Continue installing?"
                try:
                    self.ask_confirm(msg)
                except NegativeConfirmation:
                    return

            if (zip_path / INKBOARD_PACKAGE_INTERNAL_FOLDER / "platforms").exists():
                _LOGGER.info("Installing platforms")
                for platform_folder in (zip_path / INKBOARD_PACKAGE_INTERNAL_FOLDER / "platforms").iterdir():
                    
                    with suppress(NegativeConfirmation):
                        self.ask_confirm(f"Install platform {platform_folder.name}?")
                        try:
                            _LOGGER.info(f"Installing platform {platform_folder.name}")
                            self._install_platform_zipinfo(zip_file.getinfo(platform_folder.at))
                        except NegativeConfirmation:
                            pass
                        except Exception as exce:
                            _LOGGER.error(f"Could not install platform {platform_folder.name}", exc_info=exce)
                _LOGGER.info("Platforms installed")


            if (zip_path / INKBOARD_PACKAGE_INTERNAL_FOLDER / "integrations").exists():
                _LOGGER.info("Installing integrations")
                for integration_folder in (zip_path / INKBOARD_PACKAGE_INTERNAL_FOLDER / "integrations").iterdir():
                    
                    with suppress(NegativeConfirmation):
                        self.ask_confirm(f"Install integration {integration_folder.name}?")
                        try:
                            _LOGGER.info(f"Installing integration {integration_folder.name}")
                            self._install_integration_zipinfo(zip_file.getinfo(integration_folder.at))
                        except Exception as exce:
                            _LOGGER.error(f"Could not install integration {integration_folder.name}", exc_info=exce)
                _LOGGER.info("Integrations installed")

            if (zip_path / "configuration").exists():
                _LOGGER.info(f"Extracting configuration folder to current working directory {Path.cwd()}")
                ##First extract, then find requirements.txt file
                
                self.extract_zip_folder(zip_file.getinfo((zip_path / "configuration").at),
                                        allow_overwrite=True, just_contents=True)

                self.install_config_requirements(Path.cwd())

            _LOGGER.info("Package succesfully installed")
            return
        
    def install_platform(self):
        """Installs a platform type .zip file file
        """           
        file = self._file

        if self._package_type != 'platform':
            raise TypeError(f"{file} is not a platform type .zip file")
        
        with zipfile.ZipFile(file) as zip_file:
            self.__zip_file = zip_file
            zip_path = zipfile.Path(zip_file)
            p = list(zip_path.iterdir())[0]
            self._install_platform_zipinfo(zip_file.getinfo(p.at))
        
        return
    
    def install_integration(self):
        """Installs an integration type .zip file file
        """   
        
        file = self._file

        if self._package_type != 'integration':
            raise TypeError(f"{file} is not an integration type .zip file")
        
        with zipfile.ZipFile(file) as zip_file:
            self.__zip_file = zip_file
            zip_path = zipfile.Path(zip_file)
            p = list(zip_path.iterdir())[0]
            self._install_integration_zipinfo(zip_file.getinfo(p.at))
        
        return

    def install_config_requirements(self, config_file: Union[str,Path]):
        """Installs requirements for the passed config_file

        If config_file is a .yaml file it will use the folder the file is in. If is it a folder, that folder be used as a base.
        The function looks for requirements.txt files in the 'config folder' itself, in 'config folder/files' (but only the top folder), and recursively in 'config folder/custom' (i.e. it goes through all files in all folders within there and installs every requirements.txt it finds)
        Afterwards, it will look in 'config folder/custom/integrations' and install all requirements for all integrations, as well as prompt for optional requirements if a function is supplied.

        Parameters
        ----------
        config_file : Union[str,Path]
            The yaml file from which to get the base folder, or the base folder itself
        skip_confirmations : bool, optional
            Instructs pip to not prompt for confirmations when installing, by default False
        confirmation_function : Callable[[str],bool], optional
            Function to call when optional requirements can be installed, by default `confirm_input` (command line prompt). If a boolean `False` is returned, or a `NegativeConfirmation` error is raised, the optional requirements are not installed.
        """        

        skip_confirmations = self._skip_confirmations

        if isinstance(config_file,str):
            config_file = Path(config_file)
        
        if config_file.is_file():
            assert config_file.suffix in CONFIG_FILE_TYPES, "Config file must be a yaml file"
            path = config_file.parent
        else:
            path = config_file
        
        if (path / "custom").exists():
            if (path / REQUIREMENTS_FILE).exists():
                self.pip_install_requirements_file(path / REQUIREMENTS_FILE, skip_confirmations)
            
            if (path / "files" / REQUIREMENTS_FILE).exists():
                self.pip_install_packages(path / "files" / REQUIREMENTS_FILE, skip_confirmations)

            folder = path / "custom"
            for foldername, subfolders, filenames in os.walk(folder):
                if REQUIREMENTS_FILE in filenames:
                    file_path = os.path.join(foldername, REQUIREMENTS_FILE)
                    self.pip_install_requirements_file(file_path, skip_confirmations)
            
            if (folder / "integrations").exists():
                for integration_folder in (folder / "integrations").iterdir():
                    with open(integration_folder / packageidfiles["integration"]) as f:
                        integration_conf: manifestjson = json.load(f)

                    if reqs := integration_conf.get("requirements", []):
                        _LOGGER.info(f"Installing requirements for custom integration {integration_folder.name}")
                        res = self.pip_install_packages(*reqs, no_input=skip_confirmations)

                        if res.returncode != 0: 
                            _LOGGER.error(f"Something went wrong installing requirements for custom integration {integration_folder.name}")
                            continue
                    
                    for opt_req, reqs in integration_conf.get("optional_requirements", {}).items():
                        with suppress(NegativeConfirmation):
                            msg = f"Install requirements for optional features {opt_req} for custom integration {integration_folder.name}?"

                            if self._confirmation_function:
                                if not self._confirmation_function(msg, self):
                                    continue
                            self.pip_install_packages(*reqs, no_input=skip_confirmations)

    def _install_platform_zipinfo(self, platform_info: zipfile.ZipInfo):
        
        assert platform_info.is_dir(),"Platforms must be a directory"

        platform_zippath = zipfile.Path(self.__zip_file, platform_info.filename)
        platform = platform_zippath.name 

        # with self.__zip_file.open(f"{platform_info.filename}{packageidfiles['platform']}") as f:
        f = self.__zip_file.open(f"{platform_info.filename}{packageidfiles['platform']}")
        platform_conf: platformjson = json.load(f)
        platform_version = parse_version(platform_conf['version'])

        install = True
        ib_requirements = platform_conf["inkboard_requirements"]

        if not self.check_inkboard_requirements(ib_requirements, f"Platform {platform}"):
            msg = f"inkBoard requirements for platform {platform} are not met (see logs). Continue installing?"
            self.ask_confirm(msg)

        if (INKBOARD_FOLDER / "platforms" / platform).exists():
            
            with open(INKBOARD_FOLDER / "platforms" / platform / packageidfiles["platform"]) as f:
                cur_conf: platformjson = json.load(f)
                cur_version = parse_version(cur_conf['version'])
            
            if cur_version > platform_version:
                msg = f"Version {cur_version} of platform {platform} is currently installed. Do you want to install earlier version {platform_version}?"
                self.ask_confirm(msg)

            elif platform_version > cur_version:
                _LOGGER.info(f"Updating platform {platform} from version {cur_version} to {platform_version}.")
            else:
                msg = f"Version {platform_version} of platform {platform} is already installed. Do you want to overwrite it?"
                self.ask_confirm(msg)

        if not install:
            _LOGGER.info(f"Not installing platform {platform} {platform_version}")
            return
        
        _LOGGER.info(f"Installing new platform {platform}, version {platform_version}")
        if self.install_platform_requirements(platform,platform_conf):
            self.extract_zip_folder(platform_info, path = INKBOARD_FOLDER / "platforms", allow_overwrite=True)
            _LOGGER.info("Extracted platform file")
        return

    def _install_integration_zipinfo(self, integration_info: zipfile.ZipInfo):
        
        assert integration_info.is_dir(),"Integrations must be a directory"

        integration_zippath = zipfile.Path(self.__zip_file, integration_info.filename)
        integration = integration_zippath.name

        manifestpath = integration_zippath / packageidfiles['integration']
        f = manifestpath.open()
        integration_conf: manifestjson = json.load(f)
        integration_version = parse_version(integration_conf['version'])

        install = True
        ib_requirements = integration_conf.get("inkboard_requirements",{})

        if ib_requirements and not self.check_inkboard_requirements(ib_requirements, f"Integration {integration}"):
            msg = f"inkBoard requirements for integration {integration} are not met (see logs). Continue installing?"
            self.ask_confirm(msg)

        if (INKBOARD_FOLDER / "integrations" / integration).exists():
            
            with open(INKBOARD_FOLDER / "integrations" / integration / packageidfiles["integration"]) as f:
                cur_conf: manifestjson = json.load(f)
                cur_version = parse_version(cur_conf['version'])
            
            if cur_version > integration_version:
                msg = f"Version {cur_version} of Integration {integration} is currently installed. Do you want to install earlier version {integration_version}?"
                self.ask_confirm(msg)

            elif integration_version > cur_version:
                _LOGGER.info(f"Updating Integration {integration} from version {cur_version} to {integration_version}.")
            else:
                msg = f"Version {integration_version} of Integration {integration} is already installed. Do you want to overwrite it?"
                self.ask_confirm(msg)

        if not install:
            _LOGGER.info(f"Not installing Integration {integration} {integration_version}")
            return

        if self.install_integration_requirements(integration, integration_conf):
            self.extract_zip_folder(integration_info, path = INKBOARD_FOLDER / "integrations", allow_overwrite=True)
            _LOGGER.info("Extracted integration files")

    def extract_zip_folder(self, member: Union[str,zipfile.ZipInfo], path: Union[str,Path,None] = None, pwd: str = None, just_contents: bool = False, allow_overwrite: bool = False):
        """Extracts a folder and all it's contents from a ZipFile object to path.

        Parameters
        ----------
        member : zipfile.ZipInfo
            The name or ZipInfo object of the folder to extract
        path : _type_, optional
            The path to extract the folder to, by default None (which extracts to the current working directory)
        just_contents : bool
            Extract the contents of the folder directly to path, instead of extracting the folder itself, defaults to `False`
        pwd : str, optional
            Optional password for the archive, by default None
        allow_overwrite : bool
            Allows overwriting existing files or folders
        """        

        if isinstance(member, str):
            member = self.__zip_file.getinfo(member)

        assert member.is_dir(),"Member must be a directory"

        ##Gotta put it all in a temporary directory to isolate the folder correctly
        with tempfile.TemporaryDirectory() as tempdir:
            self.__zip_file.extract(member, tempdir, pwd)
            for file in self.__zip_file.namelist():
                if file.startswith(member.orig_filename) and file != member.orig_filename:
                    self.__zip_file.extract(file, tempdir, pwd)
            _LOGGER.verbose(f"Extracted folder {member.orig_filename} to temporary directory")

            if path == None:
                path = Path.cwd()
            
            if just_contents:
                src = Path(tempdir) / Path(member.orig_filename)
            else:
                src = Path(tempdir) / Path(member.orig_filename).parent

            shutil.copytree(
                src = src,
                dst = path,
                dirs_exist_ok = allow_overwrite
            )
        _LOGGER.debug("Copied from tempdir")
        ##What happens here with nested stuff? I.e. internal folders -> check with package extraction
        
        return

    @classmethod
    def gather_inkboard_packages(cls) -> dict[Path, packagetypes]:
        """Gathers all inkBoard viable packages availables in the current working directory

        Returns
        -------
        dict[Path, packagetypes]
            Dict with all the path objects of all found packages and their package type
        """

        _LOGGER.info(f"Gathering inkBoard zip packages in {Path.cwd()}")

        packs = {}
        for file in Path.cwd().glob('*.zip'):
            if file.suffix != ".zip":
                continue

            if p := cls.identify_zip_file(file):
                packs[file] = p
        
        _LOGGER.info(f"Found {len(packs)} inkBoard installable zip packages.")
        return packs
    
    @classmethod
    def identify_zip_file(cls, file: Union[str, Path, zipfile.ZipFile]) -> Optional[packagetypes]:
        """Identifies the type of inkBoard package the zip file is

        Parameters
        ----------
        file : Union[str, Path]
            The file to identify. Must be a .zip file.

        Returns
        -------
        Optional[packagetypes]
            The type of package this file is (package, integration or platform), or None if it could not be identified.

        Raises
        ------
        TypeError
            Raised if the provided file is not a zipfile
        """        


        if not isinstance(file, zipfile.ZipFile):
            zip_file = cls._path_to_zipfile(file)
            zip_file.close()
        else:
            zip_file = file
        
        p = zipfile.Path(zip_file)
        root_files = [f for f in p.iterdir()]

        if len(root_files) == 1 and root_files[0].is_dir():
            ##Look in the single folder and whether it contains a manifest or platform json
            if (root_files[0] / packageidfiles["integration"]).exists():
                return 'integration'
            elif (root_files[0] / packageidfiles["platform"]).exists():
                return 'platform'
        elif (p / packageidfiles["package"]).exists() and (len(root_files) in {2,3}):  ##2 or 3: at least contains package.json, and has .inkBoard and/or configuration folder
            return 'package'

        return

    @staticmethod
    def _path_to_zipfile(file: Union[str, Path]) -> zipfile.ZipFile:
        """Converts a path string or Path object to a `zipfile.ZipFile` object

        Parameters
        ----------
        file : Union[str, Path]
            The file to open

        Returns
        -------
        zipfile.ZipFile
            The corresponding ZipFile object

        Raises
        ------
        TypeError
            Raised if the file is not a .zip file
        """        

        if isinstance(file, str):
            file = Path(file)

        if file.suffix != '.zip':
            raise TypeError("File must be a .zip file")
        else:
            return zipfile.ZipFile(file, 'r')



class InternalInstaller(BaseInstaller):
    "Handles installing requirements already installed platforms and integrations."
    def __init__(self, install_type: internalinstalltypes, name: str, skip_confirmations = False, confirmation_function = None):
        ##May remove the subclassing, but just reuse the usable functions (i.e. seperate out a few funcs.)
        ##Also, use the constant designer mod in case something is not found internally.
        ##Do give a warning for platforms though, or integrations without a designer module.
        if install_type == "integration":
            file = Path("integrations") / name
        elif install_type == "platform":
            file = Path("platforms") / name

        full_path = INKBOARD_FOLDER / file
        if not const.DESIGNER_INSTALLED:
            assert full_path.exists(), f"{install_type} {name} is not installed or does not exist"
        else:
            if not full_path.exists():
                assert (const.DESIGNER_FOLDER / file).exists(),  f"{install_type} {name} is not installed or does not exist"
                full_path = const.DESIGNER_FOLDER / file

        self._name = full_path.name
        self._full_path = full_path
        self._confirmation_function = confirmation_function
        self._skip_confirmations = skip_confirmations
        self._install_type = install_type
        return
    
    def install(self):
        if self._install_type == "integration":
            return self.install_integration()
        elif self._install_type == "platform":
            return self.install_platform()

    def install_platform(self):

        with open(self._full_path / packageidfiles["platform"]) as f:
            conf: platformjson = json.load(f)
            
        with suppress(NegativeConfirmation):
            msg = f"Install platform {self._name}?"
            self.ask_confirm(msg)
            return self.install_platform_requirements(self._name, conf)
        return 1

    def install_integration(self):
        with open(self._full_path / packageidfiles["integration"]) as f:
            conf: platformjson = json.load(f)
        
        with suppress(NegativeConfirmation):
            msg = f"Install integration {self._name}?"
            self.ask_confirm(msg)
            return self.install_integration_requirements(self._name, conf)
        return 1



"Types for inkBoard"

from typing import TypedDict

class inkboardrequirements(TypedDict):
    "Dict for indicating inkBoard related requirements for platform.json or manifest.json"

    inkboard_version: str
    "The inkBoard version requirement. Can use a comparison if needed."

    pssm_version: str
    "Required Python Screen Stack Manager version"

    platforms: list[str]
    "List of required platforms. Requires at least one of the list entries for the requirement to be met."

    integrations: list[str]
    "List of required integrations. All integrations must be installed for the requirement to be met."


class platformjson(TypedDict):
    "Type hinting for platform.json file"

    version: str
    "Version string. Use x.x.x for versioning"

    requirements: list[str]
    "List with requirements. Follows conventions for pip install"

    optional_requirements: dict[str,list[str]]
    "Optional requirements that can be installed for i.e. implementing more features"

    inkboard_requirements: inkboardrequirements
    "inkBoard specific requirements"

    description: str
    "Optional description of the platform"

    ##Also include: maintainer etc. similar to manifest json

class manifestjson(TypedDict):
    "Dict for inidicating requirements for integrations"
    
    version: str
    "Version string. Use x.x.x for versioning"

    requirements: list[str]
    "List with requirements. Follows conventions for pip install"

    optional_requirements: dict[str,list[str]]
    "Optional requirements that can be installed for i.e. implementing more features"

    inkboard_requirements: inkboardrequirements
    "inkBoard specific requirements"


"""
Handles reading and setting up the yaml config for inkBoard
"""
import logging
from pathlib import Path
from typing import TYPE_CHECKING

from . import const

from .configure import config

if TYPE_CHECKING:
    from yaml import Node

##Will probably set up a new logger class for this that can just be passed the node
_LOGGER = logging.getLogger(__name__)

def log_yaml_line(level: int, msg: str, node: "Node", *args, **kwargs):
    yaml_line = node.start_mark.line
    yaml_file = node.start_mark.name
    msg = f"[yaml file {yaml_file} line {yaml_line}]: {msg}"
    _LOGGER.log(level, msg, *args, **kwargs)
